# -*- encoding: utf-8 -*-

from typing import Callable, Optional, TypeVar

from antlr4 import ParserRuleContext
from antlr4.error.Errors import InputMismatchException, ParseCancellationException

from ..errors import WizardError, WizardIndexError, WizardParseError, WizardTypeError

T = TypeVar("T")


def wrap_exceptions(
    fn: Callable[[], T], context: Optional[ParserRuleContext] = None
) -> T:
    """
    Wrap the given call to `fn` in a proper try/except block to convert as many
    exceptions as possible to WizardError.
    """

    try:
        return fn()
    # Just forward wizard-errors:
    except WizardError as ex:
        raise ex
    # Wrap parser exceptions:
    except ParseCancellationException as ex:
        # Try to find a context:
        if context is None and isinstance(ex.args[0], InputMismatchException):
            context = ex.args[0].ctx

        raise WizardParseError(context, ex)
    # Wrap "known" exceptions:
    except TypeError as te:
        raise WizardTypeError(context, te)
    except IndexError as ie:
        raise WizardIndexError(context, *ie.args)



# -*- encoding: utf-8 -*-

from abc import abstractmethod
from enum import Enum, auto
from typing import MutableMapping


class Severity(Enum):

    ALLOW = auto()
    WARNING = auto()
    ERROR = auto()


class Issue(Enum):

    # Indicates if multiple default values are allowed in SelectOne statement.
    MULTIPLE_DEFAULTS_IN_SELECT_ON = auto()

    # Indicates if using a non-set variable is allowed or not.
    USAGE_OF_NOTSET_VARIABLES = auto()

    # Indicates if case labels can be non-string and Identifier instead.
    USAGE_OF_NON_STRING_CASE_LABELS = auto()

    # Indicates that a Note can be anything (not only a string):
    USAGE_OF_ANYTHING_IN_NOTE = auto()

    # Select a sub-package or a plugin that does not exist:
    SELECT_MISSING_SUBPACKAGE = auto()
    SELECT_MISSING_PLUGIN = auto()


class SeverityContext:

    _default_severity: Severity
    _context: MutableMapping[Issue, Severity] = {}

    def __init__(self):
        self._default_severity = Severity.WARNING

    @abstractmethod
    def warning(self, text: str):
        """
        Display a warning.

        Args:
            text: The warning text.
        """
        ...

    def set_default_severity(self, severity: Severity):
        """
        Set the default severity level used. By default it is "WARNING".

        Args:
            severity: The new default severity level.
        """
        self._default_severity = severity

    def set_severity(self, issue: Issue, severity: Severity):
        """
        Set the severity for the given issue.

        Args:
            issue: The issue to set the severity for.
            severity: The new severity of the issue.
        """
        self._context[issue] = severity

    def raise_or_warn(self, issue: Issue, exc: Exception, warn: str):
        """
        Check the severity of the given issue and eventually raise the given exception
        or warn the user with the given message. May do nothing if the severity is
        Severity.ALLOW.

        Args:
            issue: The issue to check the severity for.
            exc: The exception to raise if the severity is "ERROR".
            warn: The message to use to warn the user if the severity is "WARNING".
        """
        s = self._context.get(issue, self._default_severity)

        if s == Severity.ERROR:
            raise exc

        if s == Severity.WARNING:
            self.warning(warn)



# -*- encoding: utf-8 -*-


from abc import abstractmethod
from typing import List, Optional


class ManagerModInterface:

    """
    The ManagerModInterface represents the interface to implement to let the
    interpreter (or runner) interact with the game information and files.
    """

    # Function methods:
    @abstractmethod
    def compareGameVersion(self, version: str) -> int:
        """
        Used to test the installed version of the game against the specified version.

        Args:
            version: A string formatted to hold a file version number, eg. "1.2.0.416".

        Returns:
            -1, 0 or 1 if the installed version is less, equal or greater than the given
            version.
        """
        ...

    @abstractmethod
    def compareSEVersion(self, version: str) -> int:
        """
        Used to test the installed version of the script extender against the
        specified version.

        Args:
            version: A string formatted to hold a file version number, eg. "0.0.20.1".

        Returns:
            -1, 0 or 1 if the installed version is less, equal or greater than the given
            version.
        """
        ...

    @abstractmethod
    def compareGEVersion(self, version: str) -> int:
        """
        Used to test the installed version of the graphics extender against the
        specified version.

        Args:
            version: A string formatted to hold a file version number, eg. "3.0.1".

        Returns:
            -1, 0 or 1 if the installed version is less, equal or greater than the given
            version.
        """
        ...

    @abstractmethod
    def compareWBVersion(self, version: str) -> int:
        """
        Used to test the installed version Wrye Bash against the
        specified version.

        Args:
            version: A string formatted to hold a file version number, eg. "307".

        Returns:
            -1, 0 or 1 if the installed version is less, equal or greater than the given
            version.
        """
        ...

    @abstractmethod
    def dataFileExists(self, *filepaths: str) -> bool:
        """
        Tests for the existence of a file in the Data directory. If the file you are
        testing for is a plugin, this should also detect ghosted versions of the file.

        Args:
            filepaths: List of files to check. Path should be relative to the data
                directory but can go up with "..".

        Returns:
            True if all the files exist, False otherwise.
        """
        ...

    @abstractmethod
    def getPluginLoadOrder(self, filename: str, fallback: int = -1) -> int:
        """
        Returns the current load order index of a plugin in the Data directory.

        Args:
            filename: Path relative to the Data direcotry to the plugin.
            fallback: Value to return if the plugin has no load order index.

        Returns:
            The load order of the plugin or fallback if the plugin has no load order
            index.
        """
        ...

    @abstractmethod
    def getPluginStatus(self, filename) -> int:
        """
        Returns the status of a plugin in the Data directory.

        Args:
            filename: Path relative to the Data direcotry to the plugin.

        Returns:
            -1 if the plugin does not exist,
            0 if the plugin is not active, imported or merged,
            1 if the plugin is not active but has portions imported into Bashed Patch,
            2 if the plugin is active,
            3 if the plugin is merged into the Bashed Patch.
        """
        ...

    @abstractmethod
    def getFilename(self, path: str) -> str:
        """
        Retrieve the filename in a path (actually from the system).

        Args:
            path: Path to the file to extract the filename from.

        Returns:
            If the path does not exists, or if the path points to a directory, returns
            an empty string, otherwise returns the filename of the file the path points
            to.
        """
        ...

    @abstractmethod
    def getFolder(self, path: str) -> str:
        """
        Retrieve the folder in a path (actually from the system).

        Args:
            path: Path to the file to extract the folder from.

        Returns:
            If the path does not exists, or if the path points to a file, returns
            an empty string, otherwise returns the name of the folder the path points
            to.
        """
        ...


class SelectOption:

    """
    Option from SelectOne or SelectMany statement.
    """

    def __init__(self, name: str, desc: str, image: Optional[str] = None):
        self._name = name
        self._desc = desc
        self._image = image

    @property
    def name(self) -> str:
        """
        Returns:
            The name of the option.
        """
        return self._name

    @property
    def description(self) -> str:
        """
        Returns:
            The description of the option.
        """
        return self._desc

    @property
    def image(self) -> Optional[str]:
        """
        Returns:
            The image for the option, or None if there are no image.
        """
        return self._image

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SelectOption):
            return False
        return (self.name, self.description, self.image) == (
            other.name,
            other.description,
            other.image,
        )

    def __repr__(self) -> str:
        return f"SelectOption({self.name}, {self.description}, {self.image})"


class ManagerUserInterface:

    """
    The ManagerUserInterface represents the interface to implement to let the
    interpreter (or runner) interact with the user.
    """

    # Development methods:
    def warning(self, text: str):
        """
        Called when a warning is emitted by the interpreter, usually due
        to an improper script.

        This should be a light warning for the user, even invisible, as it is
        mostly called for script-related stuff (e.g. use of a variable before
        it is set).
        """
        ...

    def error(self, exc: Exception):
        """
        Called when an error occurs while executing the script. Unlike warning(),
        errors cannot be recovered from and cannot be safely ignored.

        By default this method simply re-raised the given exception.

        Args:
            exc: The actual Python exception that raises the error.
        """
        raise exc

    # Choice / dialog methods:
    def cancel(self) -> bool:
        """
        Called when a 'Cancel' keyword is encountered.

        Returns:
            True to stop executing the script, False to continue.
        """
        return True

    def complete(self) -> bool:
        """
        Called when a 'Return' keyword is encountered or at the end of the
        installation.

        Returns:
            True to stop executing the script, False to continue.
        """
        return True

    @abstractmethod
    def selectOne(
        self, description: str, options: List[SelectOption], default: SelectOption
    ) -> SelectOption:
        """
        Query user to select one of the proposed options.

        Args:
            description: Description to display to the user.
            options: List of available options.
            default: Default selected option.

        Returns:
            The option selected by the user.
        """
        ...

    @abstractmethod
    def selectMany(
        self,
        description: str,
        options: List[SelectOption],
        default: List[SelectOption] = [],
    ) -> List[SelectOption]:
        """
        Query user to select many of the proposed options.

        Args:
            description: Description to display to the user.
            options: List of available options.
            default: Default selected options.

        Returns:
            The options selected by the user. Can be empty.
        """
        ...

    @abstractmethod
    def requiresVersions(
        self,
        game_version: str,
        se_version: Optional[str],
        ge_version: Optional[str],
        wrye_bash_version: Optional[str],
    ):
        """
        Tests the user system against version requirements you specify. If the
        requirements are not met, a warning dialog will be shown asking if you wish
        to continue anyway.

        Args:
            game_version: Version of the game to test for. See compareGameVersion for
                the proper format of the string.
            se_version: Version of the Script Extender to test for. See
                compareSEVersion for the proper format of the string.
            ge_version: Version of the Graphics Extender to test for. See
                compareGEVersion for the proper format of the string.
            wrye_bash_version: Version of Wrye Bash to test for. See
                compareWBVersion for more info.
        """
        ...



# -*- encoding: utf-8 -*-

from typing import Any, List, Optional


class WizardINISetting:

    _filename: str
    _section: str
    _setting: str

    def __init__(self, filename: str, section: str, setting: str):
        self._filename = filename
        self._section = section
        self._setting = setting

    @property
    def filename(self) -> str:
        """
        Returns:
            The name of the file containing the setting.
        """
        return self._filename

    @property
    def section(self) -> str:
        """
        Returns:
            The section containing the setting.
        """
        return self._section

    @property
    def setting(self) -> str:
        """
        Returns:
            The name of the setting.
        """
        return self._setting


class WizardINISettingEdit(WizardINISetting):

    _value: Any
    _comment: Optional[str]

    def __init__(
        self,
        filename: str,
        section: str,
        setting: str,
        value: Any,
        comment: Optional[str] = None,
    ):
        super().__init__(filename, section, setting)
        self._value = value
        self._comment = comment

    @property
    def value(self) -> Any:
        """
        Returns:
            The new value for the setting.
        """
        return self._value

    @property
    def comment(self) -> Optional[str]:
        """
        Returns:
            The comment for the setting.
        """
        return self._comment


class WizardINITweaks:

    _disabled: List[WizardINISetting]
    _modified: List[WizardINISettingEdit]

    def __init__(self):
        self._disabled = []
        self._modified = []

    @property
    def disabled(self) -> List[WizardINISetting]:
        """
        Returns:
            The list of INI settings that have been disabled by the script.
        """
        return self._disabled

    @property
    def modified(self) -> List[WizardINISettingEdit]:
        """
        Returns:
            The list of settings that have been created or modified by the script. May
            contain multiple edits for the same setting.
        """
        return self._modified

    def files(self) -> List[str]:
        """
        Returns:
            The list of all files containing INI tweaks (either modified or disabled).
        """
        return sorted(
            set(m.filename for m in self._modified).union(
                m.filename for m in self._disabled
            )
        )

    def tweaks(self, file: str) -> List[WizardINISetting]:
        """
        Args:
            file: The file to retrieve the tweaks for.

        Returns:
            The union of modified and disabled tweaks for the given file.
        """
        file = file.lower()
        tweaks: List[WizardINISetting] = [
            m for m in self.modified if m.filename.lower() == file
        ]
        tweaks.extend(m for m in self.disabled if m.filename.lower() == file)
        return tweaks

    def __bool__(self) -> bool:
        """
        Returns:
            True if these are any tweaks (modified or disabled settings).
        """
        return bool(self._modified) or bool(self._disabled)



# -*- encoding: utf-8 -*-

from pathlib import Path
from typing import BinaryIO, Iterable, Optional, TextIO, Union, overload

import chardet
from antlr4 import CommonTokenStream, InputStream, Parser, Token
from antlr4.error.ErrorListener import ConsoleErrorListener, ErrorListener
from antlr4.error.Errors import RecognitionException
from antlr4.error.ErrorStrategy import DefaultErrorStrategy
from antlr4.IntervalSet import IntervalSet

from .antlr4.wizardLexer import wizardLexer
from .antlr4.wizardParser import wizardParser
from .contexts import WizardInterpreterContextFactory, WizardTopLevelContext
from .contexts.utils import wrap_exceptions
from .expr import WizardExpressionVisitor
from .functions import make_basic_functions, make_manager_functions
from .keywords import WizardKeywordVisitor
from .manager import ManagerModInterface
from .runner import WizardRunnerExpressionVisitor, WizardRunnerKeywordVisitor
from .severity import SeverityContext
from .state import ContextState, WizardInterpreterState
from .value import SubPackages


class WizardErrorStrategy(DefaultErrorStrategy):

    """
    Custom error strategy that tries to recover for broken Wizard scripts.

    This class is inspired by DefaultErrorStrategy but is really experimental.
    Basically, it should recover by broken scripts by closing enclosing control
    block, e.g. if the body of a If is broken, we try to look for a Elif, Else
    or EndIf and skip everything in-between.
    """

    # "Block" contexts are context for which we need to consume the last token when
    # recovering to move to the next rule. This is basically the contexts for all the
    # rules that contains a EndXXX token at the end.
    BlockContexts = {
        wizardParser.IfStmtContext: wizardLexer.EndIf,
        wizardParser.ForStmtContext: wizardLexer.EndFor,
        wizardParser.SelectStmtContext: wizardLexer.EndSelect,
        wizardParser.WhileStmtContext: wizardLexer.EndWhile,
    }

    def recover(self, recognizer: Parser, e: RecognitionException):

        # Mark the context (and "some" parent contexts):
        recognizer._ctx.exception = e

        # Try to do custom recovery using control-flow:
        controlRecoverSet = self.getControlRecoverySet(recognizer)
        index = recognizer._input.index
        state = recognizer.state

        if controlRecoverSet.intervals:
            self.consumeUntil(recognizer, controlRecoverSet)

        # Same state, nothing was done, fallback to parent:
        if recognizer.state == state and recognizer._input.index == index:
            super().recover(recognizer, e)
        else:

            # Something was consumed, but do we have to consume the next token
            # to close the statement? Yes if the current context if a "block"
            # context and the next context its close token:
            for tctx, tok in self.BlockContexts.items():
                if (
                    isinstance(recognizer._ctx, tctx)
                    and recognizer.getTokenStream().LA(1) == tok
                ):
                    recognizer.consume()
                    break

        return recognizer.getTokenStream().LT(1)

    def getControlRecoverySet(self, recognizer: Parser):
        ctx = recognizer._ctx
        recoverSet = IntervalSet()

        while ctx is not None and ctx.invokingState >= 0:

            # If statement (for Else we only want EndIf):
            if isinstance(
                ctx, (wizardParser.IfStmtContext, wizardParser.ElifStmtContext)
            ):
                recoverSet.addOne(wizardLexer.Elif)
                recoverSet.addOne(wizardLexer.Else)
                recoverSet.addOne(wizardLexer.EndIf)
            elif isinstance(ctx, wizardParser.ElseStmtContext):
                recoverSet.addOne(wizardLexer.EndIf)

            # For/While statement:
            elif isinstance(ctx, wizardParser.ForStmtContext):
                recoverSet.addOne(wizardLexer.EndFor)
            elif isinstance(ctx, wizardParser.WhileStmtContext):
                recoverSet.addOne(wizardLexer.EndWhile)

            # Case/Default context:
            elif isinstance(
                ctx,
                (wizardParser.CaseStmtContext, wizardParser.DefaultStmtContext),
            ):
                recoverSet.addOne(wizardLexer.Break)

            # Select context:
            elif isinstance(ctx, wizardParser.SelectStmtContext):
                recoverSet.addOne(wizardLexer.EndSelect)

            ctx = ctx.parentCtx
        recoverSet.removeOne(Token.EPSILON)

        return recoverSet


def make_basic_context_factory(
    subpackages: SubPackages, severity: SeverityContext
) -> WizardInterpreterContextFactory:
    """
    Make a basic context factory from the given subpackages and severity context.
    The expression visitor includes the basic functions (see `make_basic_functions`).

    Args:
        subpackages: The subpackages used for the expression visitor.
        severity: The severity context to use.

    Returns:
        A basic context factory, built using default expression and keyword visitors.
    """
    evisitor = WizardExpressionVisitor(subpackages, make_basic_functions(), severity)
    kvisitor: WizardKeywordVisitor[WizardInterpreterState] = WizardKeywordVisitor(
        severity
    )
    return WizardInterpreterContextFactory(evisitor, kvisitor, severity)


def make_runner_context_factory(
    subpackages: SubPackages, manager: ManagerModInterface, severity: SeverityContext
) -> WizardInterpreterContextFactory:
    """
    Make a runner context factory from the given subpackages, severity context and
    manager.

    This method constructs a context factory using a `WizardRunnerExpressionVisitor`
    and a `WizardRunnerKeywordVisitor` constructed from the given parameters. The
    expression visitor also includes the basic functions (see `make_basic_functions`).

    Args:
        subpackages: The subpackages used for the expression visitor and keyword
            visitors.
        manager: The manager to use for the expression visitor (functions).
        severity: The severity context to use.

    Returns:
        A context factory, built using runner expression and keyword visitors.
    """
    functions = make_basic_functions()
    functions.update(make_manager_functions(manager, severity))
    evisitor = WizardRunnerExpressionVisitor(subpackages, functions, severity)
    kvisitor = WizardRunnerKeywordVisitor(subpackages, severity)
    return WizardInterpreterContextFactory(evisitor, kvisitor, severity)


def make_parse_wizard_context(
    script: Union[InputStream, Path, BinaryIO, TextIO, str],
    wrap_excs: bool = True,
    error_listeners: Iterable[ErrorListener] = [],
) -> wizardParser.ParseWizardContext:
    """
    Create a ParseWizardContext from the given script. Depending on the type of
    the script, the following procedure is used to parse the script:
    - If `script` is an `InputStream`, it is used as-is.
    - If `script` is a `Path`, it should point to a file containing a Wizard script.
    - If `script` is a `TextIO`, it is equivalent to the `str` version after reading
        the whole file.
    - If `script` is a `str`, an `InputStream` is constructed from it.

    Args:
        script: The script to create a context for.
        wrap_excs: If True, exceptions will be converted to Wizard exceptions (when
            possible).

    Returns:
        A ParseWizardContext extracted from the given script.
    """

    # Make a stream:
    stream: InputStream
    if isinstance(script, InputStream):
        stream = script
    elif isinstance(script, str):
        stream = InputStream(script)
    else:
        if isinstance(script, Path):
            with open(script, "rb") as fp:
                data = fp.read()
        else:
            data = script.read()  # type: ignore
        if isinstance(data, bytes):
            encoding = chardet.detect(data)["encoding"]
            text = data.decode(encoding)
        else:
            text = data
        stream = InputStream(text)

    # Create the lexer and disable console logs:
    lexer = wizardLexer(stream)
    lexer.removeErrorListener(ConsoleErrorListener.INSTANCE)

    stream = CommonTokenStream(lexer)

    # Create the parser with a custom error strategy:
    parser = wizardParser(stream)
    parser.removeErrorListener(ConsoleErrorListener.INSTANCE)
    parser._errHandler = WizardErrorStrategy()

    # Run the interpret:
    if wrap_excs:
        return wrap_exceptions(parser.parseWizard)  # type: ignore
    return parser.parseWizard()  # type: ignore


@overload
def make_top_level_context(
    script: Union[InputStream, Path, TextIO, str],
    factory: WizardInterpreterContextFactory,
    state: ContextState,
    **kwargs,
) -> WizardTopLevelContext[ContextState]:
    ...


@overload
def make_top_level_context(
    script: Union[InputStream, Path, TextIO, str],
    factory: WizardInterpreterContextFactory,
    **kwargs,
) -> WizardTopLevelContext[WizardInterpreterState]:
    ...


def make_top_level_context(
    script: Union[InputStream, Path, TextIO, str],
    factory: WizardInterpreterContextFactory,
    state: Optional[ContextState] = None,
    **kwargs,
) -> Union[
    WizardTopLevelContext[ContextState], WizardTopLevelContext[WizardInterpreterState]
]:
    """
    Create a WizardTopLevelContext from the given script.

    Args:
        script: The script to create a context for. See make_parse_wizard_context for
            details on the possible types.
        factory: The factory for the top-level context.
        state: The state for the top-level context. If None, a WizardInterpreterState()
            will be used.
        **kwargs: Extra named parameters that are passed to `make_parse_wizard_context`.

    Returns:
        A ParseWizardContext extracted from the given script.
    """

    ctx = make_parse_wizard_context(script, **kwargs)

    if state is None:
        return WizardTopLevelContext(factory, ctx, WizardInterpreterState())
    else:
        return WizardTopLevelContext(factory, ctx, state)



# -*- encoding: utf-8 -*-

from enum import Enum, auto
from pathlib import Path
from typing import Optional, TextIO, Tuple, Union

from antlr4 import InputStream

from .contexts import (
    WizardInterpreterContext,
    WizardInterpreterContextFactory,
    WizardSelectManyContext,
    WizardSelectOneContext,
    WizardTerminationContext,
    WizardTopLevelContext,
)
from .functions import make_basic_functions, make_manager_functions
from .interpreter import WizardInterpreter
from .manager import ManagerModInterface, ManagerUserInterface
from .runner import (
    WizardRunnerExpressionVisitor,
    WizardRunnerKeywordVisitor,
    WizardRunnerState,
)
from .severity import SeverityContext
from .value import SubPackages


class WizardScriptRunnerKeywordFactory(WizardRunnerKeywordVisitor):

    """
    Small extension of the runner keyword visitor to call requireVersions()
    on a manager.
    """

    _manager: ManagerUserInterface

    def __init__(
        self,
        manager: ManagerUserInterface,
        subpackages: SubPackages,
        severity: SeverityContext,
    ):
        super().__init__(subpackages, severity)
        self._manager = manager

    def visitRequireVersions(
        self,
        state: WizardRunnerState,
        game_version: str,
        script_extender_version: Optional[str],
        graphics_extender_version: Optional[str],
        wrye_bash_version: Optional[str],
    ):
        self._manager.requiresVersions(
            game_version,
            script_extender_version,
            graphics_extender_version,
            wrye_bash_version,
        )


class WizardScriptRunnerStatus(Enum):

    # A 'Cancel' instruction was encountered:
    CANCEL = auto()

    # A 'Return' instruction was encountered:
    RETURN = auto()

    # The script was completely executed:
    COMPLETE = auto()

    # An error was encountered:
    ERROR = auto()


class WizardScriptRunner(
    WizardInterpreter, ManagerModInterface, ManagerUserInterface, SeverityContext
):

    """
    The WizardRunner is a high-level interface for the interpreter that
    uses callbacks to interact with the user.

    Implementations should inherit the WizardRunner class and implements
    missing methods from the manager interfaces.
    """

    # Internal exceptions used to rewind / cancel at any point:
    class RewindFlow(Exception):

        """
        Exception used to rewind a script execution when calling rewind().
        """

        context: WizardInterpreterContext

        def __init__(self, context: WizardInterpreterContext):
            self.context = context

    class CancelFlow(Exception):

        """
        Exception used to cancel a script execution when calling abort().
        """

        ...

    # The current context:
    _ctx: WizardInterpreterContext

    def __init__(self, subpackages: SubPackages = SubPackages()):
        functions = make_basic_functions()
        functions.update(make_manager_functions(self, self))

        factory = WizardInterpreterContextFactory(
            WizardRunnerExpressionVisitor(subpackages, functions, self),
            WizardScriptRunnerKeywordFactory(self, subpackages, self),
            self,
        )

        SeverityContext.__init__(self)
        WizardInterpreter.__init__(self, factory)

    def context(self) -> WizardInterpreterContext:
        return self._ctx

    def abort(self):
        raise WizardScriptRunner.CancelFlow()

    def rewind(self, context: WizardInterpreterContext):
        raise WizardScriptRunner.RewindFlow(context)

    def make_top_level_context(  # type: ignore
        self,
        script: Union[InputStream, Path, TextIO, str],
        state: Optional[WizardRunnerState] = None,
    ) -> WizardTopLevelContext[WizardRunnerState]:
        if state is None:
            state = WizardRunnerState()
        return super().make_top_level_context(script, state)

    def run(
        self, script: Union[InputStream, Path, TextIO, str]
    ) -> Tuple[WizardScriptRunnerStatus, WizardRunnerState]:
        """
        Run the script from the given input stream.

        Args:
            script: The script to read the script from. Can be a antlr4 stream,
                a path to a script file, a string containing the script or a TextIO
                object.

        Returns:
            A tuple (result, state) from running the script where state is the last
            state of the interpreter and result the result.
        """

        # Run the interpret:
        ctx: WizardInterpreterContext = self.make_top_level_context(script)

        while True:

            self._ctx = ctx

            try:
                if isinstance(ctx, WizardSelectOneContext):
                    ctx = ctx.select(
                        self.selectOne(ctx.description, ctx.options, ctx.default)
                    )
                elif isinstance(ctx, WizardSelectManyContext):
                    ctx = ctx.select(
                        self.selectMany(ctx.description, ctx.options, ctx.defaults)
                    )
                elif isinstance(ctx, WizardTerminationContext):
                    if ctx.is_cancel():
                        if self.cancel():
                            return (WizardScriptRunnerStatus.CANCEL, ctx.state)
                    elif ctx.is_return():
                        if self.complete():
                            return (WizardScriptRunnerStatus.RETURN, ctx.state)
                    else:
                        if self.complete():
                            return (WizardScriptRunnerStatus.COMPLETE, ctx.state)

                ctx = ctx.exec()

            except WizardScriptRunner.RewindFlow as rfex:
                ctx = rfex.context

            except WizardScriptRunner.CancelFlow:
                if self.cancel():
                    return (WizardScriptRunnerStatus.CANCEL, ctx.state)

            except Exception as ex:
                self.error(ex)
                return (WizardScriptRunnerStatus.ERROR, ctx.state)

        return (WizardScriptRunnerStatus.COMPLETE, ctx.state)



# -*- encoding: utf-8 -*-

from typing import Optional

from antlr4 import Parser, ParserRuleContext, Token
from antlr4.error.Errors import (
    FailedPredicateException,
    InputMismatchException,
    NoViableAltException,
    RecognitionException,
)


class WizardError(Exception):
    """
    Base class for all Wizard errors.

    The only case where there is no context associated is if the parser fails
    at the beginning.
    """

    _ctx: ParserRuleContext

    def __init__(self, context: ParserRuleContext, *args):
        super().__init__(*args)
        self._ctx = context

    @property
    def context(self) -> Optional[ParserRuleContext]:
        return self._ctx

    @property
    def line(self) -> int:
        return self._ctx.start.line  # type: ignore

    @property
    def column(self) -> int:
        return self._ctx.start.column  # type: ignore

    # The messageXXX are taken from ANTLR4 error strategy (reportXXX):

    def escapeWSAndQuote(self, s: str) -> str:
        s = s.replace("\n", "\\n")
        s = s.replace("\r", "\\r")
        s = s.replace("\t", "\\t")
        return "'" + s + "'"

    def getTokenErrorDisplay(self, t: Token):
        if t is None:
            return "<no token>"
        s = t.text
        if s is None:
            if t.type == Token.EOF:
                s = "<EOF>"
            else:
                s = "<" + str(t.type) + ">"
        return self.escapeWSAndQuote(s)

    def messageNoViableAlternative(
        self, recognizer: Parser, e: NoViableAltException
    ) -> str:
        tokens = recognizer.getTokenStream()
        if tokens is not None:
            if e.startToken.type == Token.EOF:
                input = "<EOF>"
            else:
                input = tokens.getText(e.startToken, e.offendingToken)
        else:
            input = "<unknown input>"
        return "no viable alternative at input " + self.escapeWSAndQuote(input)

    def messageInputMismatch(
        self, recognizer: Parser, e: InputMismatchException
    ) -> str:
        msg: str = (
            "mismatched input "
            + self.getTokenErrorDisplay(e.offendingToken)
            + " expecting "
            + e.getExpectedTokens().toString(
                recognizer.literalNames, recognizer.symbolicNames
            )
        )
        return msg

    def messageFailedPredicate(self, recognizer, e) -> str:
        ruleName = recognizer.ruleNames[recognizer._ctx.getRuleIndex()]
        return f"rule {ruleName} {e.message}"

    @property
    def message(self) -> str:
        if not self.args or not isinstance(self.args[0], RecognitionException):
            return super().__str__()

        e = self.args[0]
        recognizer = e.recognizer

        if isinstance(e, NoViableAltException):
            return self.messageNoViableAlternative(recognizer, e)
        elif isinstance(e, InputMismatchException):
            return self.messageInputMismatch(recognizer, e)
        elif isinstance(e, FailedPredicateException):
            return self.messageFailedPredicate(recognizer, e)
        else:
            return f"unknown recognition error type: {type(e).__name__}"

    def __str__(self) -> str:
        return f"Line {self.line}, Column {self.column}: {self.message}"


class WizardParseError(WizardError):

    """
    Error raised when a parsing error occurs.
    """

    pass


class WizardUnsupportedOperation(WizardError):
    """
    Error raised when an operation is not supported (not yet implemented).
    """

    def __init__(self, context: ParserRuleContext, operation: str):
        super().__init__(context, f"'{operation}' operation is not implemented.")


class WizardIndexError(WizardError):

    _index: int

    def __init__(self, context: ParserRuleContext, index: int):
        super().__init__(context, f"Index {index} out of range.")
        self._index = index

    @property
    def index(self) -> int:
        return self._index


class WizardNameError(WizardError):

    _name: str

    def __init__(self, context: ParserRuleContext, name):
        super().__init__(context, f"The name '{name}' is not defined.")
        self._name = name

    @property
    def name(self) -> str:
        return self._name


class WizardTypeError(WizardError):
    pass


class WizardMissingPackageError(WizardError):

    _name: str

    def __init__(self, context: ParserRuleContext, name: str):
        super().__init__(context, f"Trying to activate missing '{name}' sub-package.")
        self._name = name

    @property
    def subpackage(self) -> str:
        return self._name


class WizardMissingPluginError(WizardError):

    _name: str

    def __init__(self, context: ParserRuleContext, name: str):
        super().__init__(context, f"Trying to activate missing '{name}' plugin.")
        self._name = name

    @property
    def subpackage(self) -> str:
        return self._name



# -*- encoding: utf-8 -*-

import operator
from abc import abstractproperty
from enum import Enum, auto
from pathlib import Path
from typing import Callable, Generic, Iterable, List, Optional, Type, TypeVar, Union


class CaseFoldNamedObject:

    _name: str

    def __init__(self, name: str):
        self._name = name

    @property
    def name(self) -> str:
        """
        Returns:
            The name of the object.
        """
        return self._name

    def _cmp(self, other: object, cmp: Callable[[str, str], bool]) -> bool:
        if isinstance(other, CaseFoldNamedObject):
            return cmp(self.name.casefold(), other.name.casefold())
        elif isinstance(other, str):
            return cmp(self.name.casefold(), other.casefold())
        else:
            raise TypeError(f"Cannot compare {type(self)} with {type(other)}.")

    def __eq__(self, other: object) -> bool:
        return self._cmp(other, operator.eq)

    def __ne__(self, other: object) -> bool:
        return self._cmp(other, operator.ne)

    def __lt__(self, other: object) -> bool:
        return self._cmp(other, operator.lt)

    def __le__(self, other: object) -> bool:
        return self._cmp(other, operator.le)

    def __gt__(self, other: object) -> bool:
        return self._cmp(other, operator.gt)

    def __ge__(self, other: object) -> bool:
        return self._cmp(other, operator.ge)

    def __hash__(self) -> int:
        return hash(self.name.casefold())

    def __str__(self) -> str:
        return self._name

    def __repr__(self) -> str:
        return repr(self._name)


class Void:

    """
    Simple class representing the result of a function call without a return
    value.
    """

    pass


class Plugin(CaseFoldNamedObject):
    ...


class SubPackage(CaseFoldNamedObject):
    @abstractproperty
    def files(self) -> Iterable[str]:
        pass

    @staticmethod
    def is_plugin(name: str) -> bool:
        """
        Check if the given name corresponds to a plugin.

        Args:
            name: The name of the file.

        Returns:
            True if the name corresponds to a plugin, False otherwise.
        """
        return name.endswith(".esp") or name.endswith(".esm") or name.endswith(".esl")

    def plugins(self) -> Iterable[Plugin]:
        return (Plugin(Path(file).name) for file in self.files if self.is_plugin(file))


class SubPackages(List[SubPackage]):

    """
    Class to wrap the 'SubPackages' variable.
    """

    ...


class VariableType(Enum):

    BOOL = auto()
    INTEGER = auto()
    FLOAT = auto()
    STRING = auto()

    VOID = auto()

    # Note: This is used only for SubPackages and item in SubPackages.
    LIST_SUBPACKAGES = auto()
    SUBPACKAGE = auto()

    @staticmethod
    def from_pytype(pytype: Type) -> "VariableType":
        if pytype is Void:
            return VariableType.VOID
        if issubclass(pytype, SubPackages):
            return VariableType.LIST_SUBPACKAGES
        if issubclass(pytype, SubPackage):
            return VariableType.SUBPACKAGE
        if pytype is bool:
            return VariableType.BOOL
        if pytype is int:
            return VariableType.INTEGER
        if pytype is float:
            return VariableType.FLOAT
        if pytype is str:
            return VariableType.STRING
        raise ValueError(f"Unknown type: {pytype}.")

    def __str__(self) -> str:
        if self is VariableType.LIST_SUBPACKAGES:
            return "SubPackage[]"

        if self is VariableType.SUBPACKAGE:
            return "SubPackage"

        return super().__str__().lower().split(".")[-1]


# Union of possible type for Wizard value:
WizardValueType = Union[bool, int, float, str, SubPackage, SubPackages, Void]


ValueType = TypeVar("ValueType", bound=WizardValueType)


class Value(Generic[ValueType]):

    """
    Represent a value of a given type, that can be a constant or the result of
    a complex expression.

    Value expose operators from the BAIN Wizard specification with proper type
    checking so you can use them directly.
    """

    _value: ValueType

    def __init__(self, value: ValueType):
        if value is None:
            value = Void()
        self._type = VariableType.from_pytype(type(value))
        self._value = value

    @property
    def type(self) -> VariableType:
        """
        Returns:
            The type of the variable.
        """
        return self._type

    def is_void(self) -> bool:
        """
        Returns:
            True if this value represent the "void" value.
        """
        return self._type == VariableType.VOID

    @property
    def value(self) -> ValueType:
        """
        Returns:
            The value of this constant.
        """
        return self._value

    def __pos__(self) -> "Value":
        if not isinstance(self._value, (int, float)):
            raise TypeError(
                f"Cannot use plus operator on variable of type {self._type}."
            )
        return Value(self._value)

    def __neg__(self) -> "Value":
        if not isinstance(self._value, (int, float)):
            raise TypeError(
                f"Cannot use minus operator on variable of type {self._type}."
            )
        return Value(-self._value)

    def __add__(self, other: "Value") -> "Value":
        if (
            not isinstance(self._value, (int, float, str))
            or not isinstance(other._value, (int, float, str))
            or isinstance(self._value, str) != isinstance(other._value, str)
        ):
            raise TypeError(f"Cannot add values of types {self._type}, {other._type}.")
        return Value(self._value + other._value)  # type: ignore

    def __sub__(self, other: "Value") -> "Value":
        if not isinstance(self._value, (int, float)) or not isinstance(
            other._value, (int, float)
        ):
            raise TypeError("Cannot substract non-numeric values.")
        return Value(self._value - other._value)

    def __mul__(self, other: "Value") -> "Value":
        if not isinstance(self._value, (int, float)) or not isinstance(
            other._value, (int, float)
        ):
            raise TypeError("Cannot multiply non-numeric values.")
        return Value(self._value * other._value)

    def __div__(self, other: "Value") -> "Value":
        if not isinstance(self._value, (int, float)) or not isinstance(
            other._value, (int, float)
        ):
            raise TypeError("Cannot divide non-numeric values.")
        return Value(self._value / other._value)

    def __pow__(self, other: "Value") -> "Value":
        if not isinstance(self._value, (int, float)) or not isinstance(
            other._value, (int, float)
        ):
            raise TypeError("Cannot raise non-numeric values.")
        return Value(self._value**other._value)

    def __mod__(self, other: "Value") -> "Value":
        if not isinstance(self._value, (int, float)) or not isinstance(
            other._value, (int, float)
        ):
            raise TypeError("Cannot modulo non-numeric values.")
        return Value(self._value % other._value)

    def logical_not(self) -> "Value":
        return Value(not self._value)

    def equals(self, other: "Value") -> "Value":
        return Value(self._value == other._value)

    def not_equals(self, other: "Value") -> "Value":
        return Value(self._value != other._value)

    def __or__(self, other: "Value") -> "Value":
        return Value(self._value or other._value)

    def __and__(self, other: "Value") -> "Value":
        return Value(self._value and other._value)

    def __gt__(self, other: "Value") -> "Value":
        if isinstance(self._value, (int, float)) and isinstance(
            other._value, (int, float)
        ):
            return Value(self._value > other._value)
        if isinstance(self._value, str) and isinstance(other._value, str):
            return Value(self._value > other._value)
        raise TypeError(f"Cannot compare values of types {self._type}, {other._type}.")

    def __ge__(self, other: "Value") -> "Value":
        if isinstance(self._value, (int, float)) and isinstance(
            other._value, (int, float)
        ):
            return Value(self._value >= other._value)
        if isinstance(self._value, str) and isinstance(other._value, str):
            return Value(self._value >= other._value)
        raise TypeError(f"Cannot compare values of types {self._type}, {other._type}.")

    def __lt__(self, other: "Value") -> "Value":
        if isinstance(self._value, (int, float)) and isinstance(
            other._value, (int, float)
        ):
            return Value(self._value < other._value)
        if isinstance(self._value, str) and isinstance(other._value, str):
            return Value(self._value < other._value)
        raise TypeError(f"Cannot compare values of types {self._type}, {other._type}.")

    def __le__(self, other: "Value") -> "Value":
        if isinstance(self._value, (int, float)) and isinstance(
            other._value, (int, float)
        ):
            return Value(self._value <= other._value)
        if isinstance(self._value, str) and isinstance(other._value, str):
            return Value(self._value <= other._value)
        raise TypeError(f"Cannot compare values of types {self._type}, {other._type}.")

    def contains(self, item: "Value", case_insensitive: bool = False) -> "Value":
        if not isinstance(self._value, (SubPackage, SubPackages)):
            raise TypeError(f"Cannot iterate variable of type {self._type}.")

        if not isinstance(item._value, (str, SubPackage)):
            raise TypeError(f"Cannot check presence of variable of type {self._type}.")

        if isinstance(item._value, SubPackage):
            item = Value(item._value.name)

        if case_insensitive:
            item = Value(item._value.lower())

        it: Iterable[str]
        if isinstance(self._value, SubPackages):
            it = (s.name for s in self._value)
        else:
            it = self._value.files

        for istr in it:
            if case_insensitive:
                istr = istr.lower()
            if item._value == istr:
                return Value(True)

        return Value(False)

    def __getitem__(self, index: "Value") -> "Value":
        if not isinstance(self._value, (str, SubPackage, SubPackages)):
            raise TypeError(f"Cannot index variable of type {self._type}.")
        if not isinstance(index._value, (int)):
            raise TypeError(f"Cannot index with variable of type {index._type}.")

        value: Union[str, SubPackages]
        if isinstance(self._value, SubPackage):
            value = self._value.name
        else:
            value = self._value

        try:
            return Value(value[index._value])
        except IndexError:
            raise IndexError(index._value)

    def slice(
        self, start: Optional["Value"], end: Optional["Value"], step: Optional["Value"]
    ) -> "Value":

        if not isinstance(self._value, str):
            raise TypeError(f"Cannot access slice of variable of type {self._type}.")

        for v in (start, end, step):
            if v is not None and not isinstance(v.value, int):
                raise TypeError(f"Cannot index with variable of type {v.type}.")

        return Value(
            self._value[
                slice(*(None if v is None else v.value for v in (start, end, step)))
            ]
        )

    def __iter__(self) -> Iterable["Value"]:
        if not isinstance(self._value, (SubPackage, SubPackages)):
            raise TypeError(f"Cannot iterate variable of type {self._type}.")

        it: Iterable[str]
        if isinstance(self._value, SubPackages):
            it = (s.name for s in self._value)
        else:
            it = self._value.files

        return (Value(x) for x in it)

    # Those operations are not "Wizardly", i.e. they make sense in Python:

    def __eq__(self, other: object) -> bool:
        value: Value
        if not isinstance(other, Value):
            value = Value(other)  # type: ignore
        else:
            value = other

        if self.is_void() and value.is_void():
            return True

        return value._type == self._type and value._value == self._value

    def __ne__(self, other: object) -> bool:
        return not (self == other)

    def __bool__(self) -> bool:
        return bool(self._value)

    def __repr__(self) -> str:
        return "{}({})".format(self.type, self.value)



# -*- encoding: utf-8 -*-

from pathlib import Path
from typing import Any, Optional, Sequence, TextIO, Type, Union, overload

from antlr4 import InputStream

from .contexts import (
    WizardInterpreterContext,
    WizardInterpreterContextFactory,
    WizardTerminationContext,
    WizardTopLevelContext,
)
from .state import ContextState, WizardInterpreterState
from .utils import make_top_level_context


class WizardInterpreter:

    """
    The WizardInterpreter is the main interpreter for Wizard scripts.
    """

    # The severity context:
    _factory: WizardInterpreterContextFactory

    def __init__(
        self,
        factory: WizardInterpreterContextFactory,
    ):
        """
        Args:
            factory: The context factory.
        """
        self._factory = factory

    def exec_until(
        self,
        context: WizardInterpreterContext[ContextState, Any],
        targets: Sequence[Type[WizardInterpreterContext]],
    ) -> WizardInterpreterContext[ContextState, Any]:
        """
        Execute the contexts until a context of the given type is found, or
        a termination context is found. If context is one of the given targets,
        the function returns immediately.

        Args:
            context: The start context.
            targets: Target context types. If one of those context is found,
                the process stops.

        Returns:
            The next context (one of the targets or a termination contexts).
        """
        stargets = tuple(targets)

        while not isinstance(context, stargets + (WizardTerminationContext,)):
            context = context.exec()

        return context

    @overload
    def make_top_level_context(
        self, script: Union[InputStream, Path, TextIO, str], state: ContextState
    ) -> WizardTopLevelContext[ContextState]:
        ...

    @overload
    def make_top_level_context(
        self, script: Union[InputStream, Path, TextIO, str]
    ) -> WizardTopLevelContext[WizardInterpreterState]:
        ...

    def make_top_level_context(
        self,
        script: Union[InputStream, Path, TextIO, str],
        state: Optional[ContextState] = None,
    ) -> Union[
        WizardTopLevelContext[ContextState],
        WizardTopLevelContext[WizardInterpreterState],
    ]:
        return make_top_level_context(script, self._factory, state)  # type: ignore



# -*- encoding: utf-8 -*-

from typing import Any, List, Mapping, MutableMapping, Optional, Sequence, cast

from .contexts import WizardKeywordContext
from .errors import WizardMissingPackageError, WizardMissingPluginError
from .expr import WizardExpressionVisitor
from .keywords import WizardKeywordVisitor
from .severity import Issue, SeverityContext
from .state import WizardInterpreterState
from .tweaks import WizardINISetting, WizardINISettingEdit, WizardINITweaks
from .value import Plugin, SubPackage, SubPackages


class WizardRunnerState(WizardInterpreterState):

    """
    Wrapper around multiple containers that are updated during the execution
    of a Wizard script and need to be rewound.
    """

    # The list of selected subpackages and plugins:
    _subpackages: List[SubPackage]
    _plugins: List[Plugin]

    # Renaming of plugins (original name -> new name):
    _renames: MutableMapping[Plugin, str]

    # The INI tweaks (disabled and modified settings):
    _tweaks: WizardINITweaks

    # The list of notes:
    _notes: List[str]

    def __init__(self):
        super().__init__()
        self._subpackages = []
        self._plugins = []
        self._renames = {}
        self._tweaks = WizardINITweaks()
        self._notes = []

    def copy(self) -> "WizardRunnerState":
        state: WizardRunnerState = super().copy()  # type: ignore
        state._subpackages.extend(self._subpackages)
        state._plugins.extend(self._plugins)
        state._renames.update(self._renames)
        state._notes.extend(self._notes)
        state._tweaks._disabled.extend(self.tweaks.disabled)
        state._tweaks._modified.extend(self.tweaks.modified)
        return state

    @property
    def subpackages(self) -> Sequence[SubPackage]:
        """
        Returns:
            The name of the selected sub-packages (sorted).
        """
        return [sp for sp in sorted(self._subpackages)]

    @property
    def plugins(self) -> Sequence[Plugin]:
        """
        Returns:
            The name of the selected plugins (sorted).
        """
        return [sp for sp in sorted(self._plugins)]

    @property
    def renames(self) -> Mapping[Plugin, str]:
        """
        Returns:
            The mapping for renamed plugins (original name -> new name).
        """
        return self._renames

    @property
    def tweaks(self) -> WizardINITweaks:
        """
        Returns:
            The INI tweaks created by the script (disabled, new or
            modified settings).
        """
        return self._tweaks

    @property
    def notes(self) -> Sequence[str]:
        """
        Returns:
            The list of notes.
        """
        return self._notes


class WizardRunnerExpressionVisitor(WizardExpressionVisitor):

    """
    Simple extension of the expression visitor to update INI tweaks.
    """

    def visitDisableINILine(
        self, state: WizardInterpreterState, filename: str, section: str, setting: str
    ):
        """
        Create an INI tweak file that disables the specified setting by commenting it
        out. Otherwise, behaves identically to editINI.

        Args:
            filename: The name of the INI file to edit, relative to the Data directory.
            section: The section in the INI where setting resides.
            setting: The setting to disable.
        """
        assert isinstance(state, WizardRunnerState)
        state.tweaks._disabled.append(WizardINISetting(filename, section, setting))

    def visitEditINI(
        self,
        state: WizardInterpreterState,
        filename: str,
        section: str,
        setting: str,
        value: Any,
        comment: Optional[str] = "",
    ):
        """
        Create an INI tweak file with some tweaks in it. If file that to apply the
        tweak to is from the current installer or is the game's ini file, then the
        tweaks are also applied, otherwise, it will just be generated for the user
        to apply manually.

        Args:
            filename: The name of the INI file to edit, relative to the Data directory.
            section: The section in the INI where setting resides.
            setting: The setting to set.
            value: The value to set the setting to.
            comment: An optional comment to add.
        """
        assert isinstance(state, WizardRunnerState)
        state.tweaks._modified.append(
            WizardINISettingEdit(filename, section, setting, value, comment)
        )


class WizardRunnerKeywordVisitor(WizardKeywordVisitor):

    _subpackages: SubPackages
    _plugins: List[Plugin]

    def __init__(self, subpackages: SubPackages, severity: SeverityContext):
        super().__init__(severity)
        self._subpackages = subpackages

        # use a dictionary to keep insertion order and remove duplicates
        plugins = {pg: True for sp in self._subpackages for pg in sp.plugins()}
        self._plugins = list(plugins.keys())

    @property
    def subpackages(self) -> SubPackages:
        """
        Returns:
            The list of all available subpackages.
        """
        return self._subpackages

    @property
    def plugins(self) -> List[Plugin]:
        """
        Returns:
            The list of all available plugins.
        """
        return self._plugins

    def plugins_for(self, subpackage: SubPackage) -> List[Plugin]:
        """
        Returns:
            The list of plugin names in the given subpackage.
        """
        return list(subpackage.plugins())

    def visitDeSelectAll(self, context: WizardKeywordContext, state: WizardRunnerState):
        state._subpackages.clear()
        state._plugins.clear()

    def visitDeSelectAllPlugins(
        self, context: WizardKeywordContext, state: WizardRunnerState
    ):
        """
        Args:
            state: The interpreter state to update.
        """
        state._plugins.clear()

    def visitDeSelectPlugin(
        self, context: WizardKeywordContext, state: WizardRunnerState, name: str
    ):
        """
        Args:
            state: The interpreter state to update.
            name: The name of the plugin to de-select.
        """
        if name in state._plugins:
            state._plugins.remove(cast(Plugin, name))

    def visitDeSelectSubPackage(
        self, context: WizardKeywordContext, state: WizardRunnerState, name: str
    ):
        """
        Args:
            state: The interpreter state to update.
            name: The name of the subpackage to de-select.
        """
        if name in state._subpackages:
            state._subpackages.remove(cast(SubPackage, name))

    def visitNote(
        self, context: WizardKeywordContext, state: WizardRunnerState, text: str
    ):
        """
        Args:
            state: The interpreter state to update.
            text: The text for the note.
        """
        state._notes.append(text)

    def visitRenamePlugin(
        self,
        context: WizardKeywordContext,
        state: WizardRunnerState,
        original_name: str,
        new_name: str,
    ):
        """
        Args:
            state: The interpreter state to update.
            original_name: The original name of the plugin.
            new_name: The new name of the plugin.
        """
        state._renames[Plugin(original_name)] = new_name

    def visitRequireVersions(
        self,
        state: WizardRunnerState,
        game_version: str,
        script_extender_version: Optional[str],
        graphics_extender_version: Optional[str],
        wrye_bash_version: Optional[str],
    ):
        """
        Args:
            state: The interpreter state to update.
            game_version: The required game version.
            script_extender_version: The required script extender version.
            graphics_extender_version: The required graphics extender version.
            wrye_bash_version: The required wrye bash version.
        """
        ...

    def visitResetAllPluginNames(
        self, context: WizardKeywordContext, state: WizardRunnerState
    ):
        """
        Args:
            state: The interpreter state to update.
        """
        state._renames.clear()

    def visitResetPluginName(
        self, context: WizardKeywordContext, state: WizardRunnerState, name: str
    ):
        """
        Args:
            state: The interpreter state to update.
            name: The original name of the plugin.
        """
        if name in state._renames:
            del state._renames[cast(Plugin, name)]

    def visitSelectAll(self, context: WizardKeywordContext, state: WizardRunnerState):
        """
        Args:
            state: The interpreter state to update.
        """
        state._subpackages = list(self._subpackages)
        self.visitSelectAllPlugins(context, state)

    def visitSelectAllPlugins(
        self, context: WizardKeywordContext, state: WizardRunnerState
    ):
        """
        Args:
            state: The interpreter state to update.
        """
        # Guess we only select plugins from selected subpackages?
        state._plugins = [pg for sp in state._subpackages for pg in sp.plugins()]

    def visitSelectPlugin(
        self, context: WizardKeywordContext, state: WizardRunnerState, name: str
    ):
        """
        Args:
            state: The interpreter state to update.
            name: The name of the plugin to select.
        """
        try:
            ipg = self._plugins.index(cast(Plugin, name))
        except ValueError:
            self._severity.raise_or_warn(
                Issue.SELECT_MISSING_PLUGIN,
                WizardMissingPluginError(context.context, name),
                f"Trying to select plugin '{name}' that does not exist.",
            )
            return

        state._plugins.append(self._plugins[ipg])

    def visitSelectSubPackage(
        self, context: WizardKeywordContext, state: WizardRunnerState, name: str
    ):
        """
        Args:
            state: The interpreter state to update.
            name: The name of the subpackage to select.
        """

        # Find the package:
        try:
            isp = self._subpackages.index(name)  # type: ignore
        except ValueError:
            self._severity.raise_or_warn(
                Issue.SELECT_MISSING_SUBPACKAGE,
                WizardMissingPackageError(context.context, name),
                f"Trying to select sub-package '{name}' that does not exist.",
            )
            return

        subpackage = self._subpackages[isp]
        state._subpackages.append(subpackage)

        # Auto-select plugins?
        for plugin in subpackage.plugins():
            state._plugins.append(plugin)



# -*- encoding: utf-8 -*-

from typing import Iterator, MutableMapping, Optional, Set, TypeVar

K = TypeVar("K")
V = TypeVar("V")


class StackMutableMapping(MutableMapping[K, V]):

    """
    Class representing a stacked mapping. It exposes the usual MutableMapping
    interface, but contains a "fallback" mapping and expose the union of both.
    """

    _parent: Optional["StackMutableMapping[K, V]"]
    _variables: MutableMapping[K, V]
    _deleted: Set[K]

    def __init__(self, parent: Optional["StackMutableMapping[K, V]"] = None):
        self._parent = parent
        self._variables = {}
        self._deleted = set()

    def __delitem__(self, item: K):
        # Remove the item from the list of variables:
        if item in self._variables:
            del self._variables[item]

        # If item is not in the parent:
        elif self._parent is None or item not in self._parent:
            raise KeyError(item)

        self._deleted.add(item)

    def __setitem__(self, item: K, value: V):
        # Remove the item from the list of deleted item:
        if item in self._deleted:
            self._deleted.remove(item)

        # Set the value:
        self._variables[item] = value

    def __getitem__(self, item: K) -> V:
        # If item in the current list, returns it:
        if item in self._variables:
            return self._variables[item]

        # If item was deleted, throw:
        if item in self._deleted or self._parent is None:
            raise KeyError(item)

        # Fallback to parent:
        return self._parent[item]

    def __iter__(self) -> Iterator[K]:
        # Retrieve the parent keys:
        keys: Set[K] = set()

        if self._parent is not None:
            keys.update(self._parent)
            keys.difference_update(self._deleted)

        keys.update(self._variables)

        return iter(keys)

    def __len__(self) -> int:
        # Retrieve the parent keys:
        keys: Set[K] = set()

        if self._parent is not None:
            keys.update(self._parent)
            keys.difference_update(self._deleted)

        keys.update(self._variables)

        return len(keys)

    def __repr__(self):
        return "StackMutableMapping({})".format({k: self[k] for k in self})



# Generated from ./vendor/wizparse/wizards/wizard.g4 by ANTLR 4.10.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .wizardParser import wizardParser
else:
    from wizardParser import wizardParser

# This class defines a complete generic visitor for a parse tree produced by wizardParser.

class wizardVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by wizardParser#parseWizard.
    def visitParseWizard(self, ctx:wizardParser.ParseWizardContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#body.
    def visitBody(self, ctx:wizardParser.BodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#stmt.
    def visitStmt(self, ctx:wizardParser.StmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#assignment.
    def visitAssignment(self, ctx:wizardParser.AssignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#compoundAssignment.
    def visitCompoundAssignment(self, ctx:wizardParser.CompoundAssignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#Break.
    def visitBreak(self, ctx:wizardParser.BreakContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#Cancel.
    def visitCancel(self, ctx:wizardParser.CancelContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#Continue.
    def visitContinue(self, ctx:wizardParser.ContinueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#For.
    def visitFor(self, ctx:wizardParser.ForContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#If.
    def visitIf(self, ctx:wizardParser.IfContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#Return.
    def visitReturn(self, ctx:wizardParser.ReturnContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#Select.
    def visitSelect(self, ctx:wizardParser.SelectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#While.
    def visitWhile(self, ctx:wizardParser.WhileContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#cancelStmt.
    def visitCancelStmt(self, ctx:wizardParser.CancelStmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#caseStmt.
    def visitCaseStmt(self, ctx:wizardParser.CaseStmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#defaultStmt.
    def visitDefaultStmt(self, ctx:wizardParser.DefaultStmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#elifStmt.
    def visitElifStmt(self, ctx:wizardParser.ElifStmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#elseStmt.
    def visitElseStmt(self, ctx:wizardParser.ElseStmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#forStmt.
    def visitForStmt(self, ctx:wizardParser.ForStmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#forRangeHeader.
    def visitForRangeHeader(self, ctx:wizardParser.ForRangeHeaderContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#forInHeader.
    def visitForInHeader(self, ctx:wizardParser.ForInHeaderContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#ifStmt.
    def visitIfStmt(self, ctx:wizardParser.IfStmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#selectStmt.
    def visitSelectStmt(self, ctx:wizardParser.SelectStmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#selectCaseList.
    def visitSelectCaseList(self, ctx:wizardParser.SelectCaseListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#optionTuple.
    def visitOptionTuple(self, ctx:wizardParser.OptionTupleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#selectOne.
    def visitSelectOne(self, ctx:wizardParser.SelectOneContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#selectMany.
    def visitSelectMany(self, ctx:wizardParser.SelectManyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#whileStmt.
    def visitWhileStmt(self, ctx:wizardParser.WhileStmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#argList.
    def visitArgList(self, ctx:wizardParser.ArgListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#keywordStmt.
    def visitKeywordStmt(self, ctx:wizardParser.KeywordStmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#PreIncrement.
    def visitPreIncrement(self, ctx:wizardParser.PreIncrementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#DotFunctionCall.
    def visitDotFunctionCall(self, ctx:wizardParser.DotFunctionCallContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#Or.
    def visitOr(self, ctx:wizardParser.OrContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#In.
    def visitIn(self, ctx:wizardParser.InContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#PostDecrement.
    def visitPostDecrement(self, ctx:wizardParser.PostDecrementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#TimesDivideModulo.
    def visitTimesDivideModulo(self, ctx:wizardParser.TimesDivideModuloContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#Index.
    def visitIndex(self, ctx:wizardParser.IndexContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#Exponentiation.
    def visitExponentiation(self, ctx:wizardParser.ExponentiationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#PlusMinus.
    def visitPlusMinus(self, ctx:wizardParser.PlusMinusContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#Not.
    def visitNot(self, ctx:wizardParser.NotContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#Lesser.
    def visitLesser(self, ctx:wizardParser.LesserContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#Negative.
    def visitNegative(self, ctx:wizardParser.NegativeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#Equal.
    def visitEqual(self, ctx:wizardParser.EqualContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#And.
    def visitAnd(self, ctx:wizardParser.AndContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#Slice.
    def visitSlice(self, ctx:wizardParser.SliceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#Value.
    def visitValue(self, ctx:wizardParser.ValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#Greater.
    def visitGreater(self, ctx:wizardParser.GreaterContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#FunctionCall.
    def visitFunctionCall(self, ctx:wizardParser.FunctionCallContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#PostIncrement.
    def visitPostIncrement(self, ctx:wizardParser.PostIncrementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#ParenExpr.
    def visitParenExpr(self, ctx:wizardParser.ParenExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#PreDecrement.
    def visitPreDecrement(self, ctx:wizardParser.PreDecrementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#constant.
    def visitConstant(self, ctx:wizardParser.ConstantContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#integer.
    def visitInteger(self, ctx:wizardParser.IntegerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#decimal.
    def visitDecimal(self, ctx:wizardParser.DecimalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by wizardParser#string.
    def visitString(self, ctx:wizardParser.StringContext):
        return self.visitChildren(ctx)



del wizardParser


# -*- encoding: utf-8 -*-

"""
This file contains the implementation of the "basic" functions for BAIN Wizard,
i.e. functions that do not require specific handling.
"""

import inspect
from typing import Callable, Dict, List, Mapping, Union

from .expr import SubPackage, SubPackages, Value, VariableType, Void
from .manager import ManagerModInterface
from .severity import SeverityContext
from .state import WizardInterpreterState


class WizardFunctions:

    """
    This is simply a class containing all the basic functions.
    """

    def str(self, value: Value) -> Value:
        return Value(str(value.value))

    def int(self, value: Value) -> Value:
        if isinstance(value.value, (SubPackage, SubPackages, Void)):
            return Value(0)
        return Value(int(value.value))

    def float(self, value: Value) -> Value:
        if isinstance(value.value, (SubPackage, SubPackages, Void)):
            return Value(0.0)
        return Value(float(value.value))

    def len(self, value: Value) -> Value:
        if not isinstance(value.value, str):
            return Value(0)
        return Value(len(value.value))

    def startswith(self, value: Value, *args: Value) -> Value:
        if not isinstance(value.value, str):
            return Value(False)
        for prefix in args:
            if not isinstance(prefix.value, str):
                continue
            if value.value.startswith(prefix.value):
                return Value(True)
        return Value(False)

    def endswith(self, value: Value, *args: Value) -> Value:
        if not isinstance(value.value, str):
            return Value(False)
        for prefix in args:
            if not isinstance(prefix.value, str):
                continue
            if value.value.endswith(prefix.value):
                return Value(True)
        return Value(False)

    def lower(self, value: Value) -> Value:
        if not isinstance(value.value, str):
            return value
        return Value(value.value.lower())

    def find(
        self,
        string: Value,
        substring: Value,
        start: Value = Value(0),
        end: Value = Value(-1),
    ) -> Value:
        if (
            not isinstance(string.value, str)
            or not isinstance(substring.value, str)
            or not isinstance(start.value, int)
            or not isinstance(end.value, int)
        ):
            return Value(-1)

        return Value(
            string.value.find(
                substring.value, start.value, end.value if end.value != -1 else None
            )
        )

    def rfind(
        self,
        string: Value,
        substring: Value,
        start: Value = Value(0),
        end: Value = Value(-1),
    ) -> Value:
        if (
            not isinstance(string.value, str)
            or not isinstance(substring.value, str)
            or not isinstance(start.value, int)
            or not isinstance(end.value, int)
        ):
            return Value(-1)

        return Value(
            string.value.rfind(
                substring.value, start.value, end.value if end.value != -1 else None
            )
        )


class optional:
    t: type

    def __init__(self, t: type):
        self.t = t


def wrap_function(
    name: str,
    method,
    *args: Union[optional, type],
    varargs: bool = False,
    rawargs: bool = False,
) -> Callable[[WizardInterpreterState, List[Value]], Value]:

    """
    Wrap the given function to be usable by the Wizard expression visitor.

    Args:
        name: The name of the function, for logging purpose (warning / errors).
        method: The function to wrap.
        *args: The type of arguments expected by the method.
        varargs: True if the method accept any number of arguments.
        rawargs: True if the method accept Value(), False to extract the underlying
            object.
    """

    def fn(st: WizardInterpreterState, vs: List[Value]) -> Value:

        # List of Python arguments:
        pargs: List = []

        if not varargs and len(vs) > len(args):
            raise TypeError(f"{name}: too many arguments.")

        for iarg, arg in enumerate(args):
            if iarg >= len(vs) and not isinstance(arg, optional):
                raise TypeError(f"{name}: missing required positional argument(s).")

            tp: type
            if isinstance(arg, optional):
                tp = arg.t
            else:
                tp = arg

            if not isinstance(vs[iarg].value, tp):
                raise TypeError(
                    f"Argument at position {iarg} has incorrect type for"
                    f" {name}, expected {VariableType.from_pytype(tp)} got"
                    f" {vs[iarg].type}."
                )

            if rawargs:
                pargs.append(vs[iarg])
            else:
                pargs.append(vs[iarg].value)

        ret: Value = method(*pargs)
        if ret is None:
            ret = Void()

        if not rawargs:
            ret = Value(ret)  # type: ignore

        return ret

    return fn


def make_basic_functions() -> Dict[
    str, Callable[[WizardInterpreterState, List[Value]], Value]
]:
    """
    Create a list of basic functions.

    Args:
        wf: The WizardFunctions object to use for the basic functions.

    Returns:
        A mapping from function name to basic functions, including methods.
    """

    wf = WizardFunctions()
    fns: Dict[str, Callable[[WizardInterpreterState, List[Value]], Value]] = {}

    # Add all the free functions:
    for fname in dir(wf):
        if fname.startswith("__"):
            continue
        sig = inspect.signature(getattr(wf, fname))
        varargs = "args" in sig.parameters
        fns[fname] = wrap_function(
            fname,
            getattr(wf, fname),
            *(object for _ in range(len(sig.parameters) - varargs)),
            varargs=varargs,
            rawargs=True,
        )

    # Add methods:
    for fname in dir(wf):
        if fname.startswith("__"):
            continue
        fns["string." + fname] = fns[fname]

    for t in ("integer", "float", "bool"):
        fns[t + ".str"] = fns["str"]

    return fns


def make_manager_functions(
    manager: ManagerModInterface, scontext: SeverityContext
) -> Mapping[str, Callable[[WizardInterpreterState, List[Value]], Value]]:
    """
    Add manager functions to the _functions member variables. These functions
    calls method of the manager passed in.

    Args:
        manager: The manager to delete the call to.
        scontext: The severity context to use (for the Note keyword for instance).
    """

    fns: Dict[str, Callable[[WizardInterpreterState, List[Value]], Value]] = {}

    for t in [
        # Functions:
        ("CompareGameVersion", manager.compareGameVersion, str),
        ("CompareSEVersion", manager.compareSEVersion, str),
        ("CompareGEVersion", manager.compareGEVersion, str),
        ("CompareWBVersion", manager.compareWBVersion, str),
        ("GetPluginLoadOrder", manager.getPluginLoadOrder, str, optional(int)),
        ("GetPluginStatus", manager.getPluginStatus, str),
        ("GetEspmStatus", manager.getPluginStatus, str),
        ("GetFilename", manager.getFilename, str),
        ("GetFolder", manager.getFolder, str),
    ]:
        fns[t[0]] = wrap_function(*t)  # type: ignore

    # Varargs:
    fns["DataFileExists"] = wrap_function(
        "DataFileExists", manager.dataFileExists, str, varargs=True
    )

    return fns



# -*- encoding: utf-8 -*-

"""
This file contains "context" which are used to rewind the
interpreter.
"""

from abc import abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

from antlr4 import ParserRuleContext

from ..antlr4.wizardParser import wizardParser
from ..errors import WizardError, WizardNameError, WizardParseError, WizardTypeError
from ..manager import SelectOption
from ..severity import Issue
from ..state import ContextState
from ..value import SubPackage, SubPackages, Value
from .utils import wrap_exceptions

if TYPE_CHECKING:
    from .factory import WizardInterpreterContextFactory


RuleContext = TypeVar("RuleContext", bound=ParserRuleContext)


class WizardInterpreterContext(Generic[ContextState, RuleContext]):

    """
    The base context class. The context are represented by a tree, where each
    child has access to his parent. Unlike ANTLR4 context, not all rules or
    elements in the grammer creates a context.
    """

    # The interpreter:
    _factory: "WizardInterpreterContextFactory"

    # The ANTLR4 context:
    _context: RuleContext

    # The parent context:
    _parent: "WizardInterpreterContext"

    # The state of this context:
    _state: ContextState

    def __init__(
        self,
        factory: "WizardInterpreterContextFactory",
        context: RuleContext,
        parent: "WizardInterpreterContext[ContextState, Any]",
        state: Optional[ContextState] = None,
    ):
        """
        Args:
            factory: The context factory.
            context: The ANTLR4 context.
            parent: The parent context.
            state: The state for this context. If None, a copy of the
                parent state will be made.
        """
        self._factory = factory
        self._context = context
        self._parent = parent

        if state is None:
            self._state = parent.state.copy()  # type: ignore
        else:
            self._state = state

    @property
    def factory(self) -> "WizardInterpreterContextFactory":
        """
        Returns:
            The factory associated with this context.
        """
        return self._factory

    @property
    def parent(self) -> "WizardInterpreterContext":
        """
        Returns:
            The parent context.
        """
        return self._parent

    @property
    def context(self) -> RuleContext:
        """
        Returns:
            The ANTLR4 context associated with this interpreter context.
        """
        return self._context

    @property
    def state(self) -> ContextState:
        """
        Returns:
            The state before the execution of this context.
        """
        return self._state

    @abstractmethod
    def exec_(self) -> "WizardInterpreterContext":
        """
        Execute the next 'step' of this context and returns the next context
        to execute. When done, should return the parent context.

        This method should be implement by all sub-classes representing specific
        contexts.
        """
        pass

    def exec(self) -> "WizardInterpreterContext":
        """
        Execute the next 'step' of this context and returns the next context
        to execute. When done, should return the parent context.
        """
        if self.context.exception:
            raise WizardParseError(self.context, self.context.exception)
        return wrap_exceptions(self.exec_, self.context)


class WizardBreakableContext:

    """
    Context that can be broken by a 'Break' expression.
    """

    @abstractmethod
    def break_(self) -> WizardInterpreterContext:
        """
        Called when a 'Break' instruction is encountered, should usually return
        the parent.

        Returns:
            The next context to use after this break.
        """
        pass


class WizardContinuableContext:

    """
    Context that can be broken by a 'Continue' expression.
    """

    @abstractmethod
    def continue_(self) -> WizardInterpreterContext:
        """
        Called when a 'Continue' instruction is encountered, should usually return
        the context itself.

        Returns:
            The next context to use after this continue.
        """
        pass


class WizardTerminationContext(
    WizardInterpreterContext[ContextState, wizardParser.ParseWizardContext]
):

    """
    Special context returned at the end of the execution. You can check what caused
    the termination using the `is_cancel`, `is_return` and `is_complete` method. If you
    want to check for "normal" completion, you can check for `not is_return()`.
    """

    def exec_(self) -> WizardInterpreterContext:
        raise NotImplementedError("exec() should not be called on termination context.")

    def is_cancel(self) -> bool:
        """
        Returns:
            True if this termination was caused by a 'Cancel' keyword.
        """
        return isinstance(self.parent, WizardCancelContext)

    def message(self) -> Optional[str]:
        """
        Returns:
            The cancel message, if any. If `is_cancel()` is False, this returns None.
        """
        if not isinstance(self.parent, WizardCancelContext):
            return None
        return self.parent.message()

    def is_return(self) -> bool:
        """
        Returns:
            True if this termination was caused by a 'Return' keyword.
        """
        return isinstance(self.parent, WizardReturnContext)

    def is_complete(self) -> bool:
        """
        Returns:
            True if this termination was caused by the end of the script.
        """
        return not isinstance(self.parent, (WizardCancelContext, WizardReturnContext))


class WizardTopLevelContext(
    WizardInterpreterContext[ContextState, wizardParser.ParseWizardContext]
):

    """
    Special context to be the entry context point.
    """

    def __init__(
        self,
        factory: "WizardInterpreterContextFactory",
        context: wizardParser.ParseWizardContext,
        state: ContextState,
        parent: Optional[WizardInterpreterContext] = None,
    ):
        """
        Args:
            factory: The factory used to create this context and future contexts.
            context: The ANTLR4 context.
            state: The starting state for this context.
            parent: If not None, it means that this context was created from an Exec()
                expression and the parent should be the enclosing body.
        """
        super().__init__(factory, context, self if parent is None else parent, state)

    def exec_(self) -> WizardInterpreterContext:
        # We forward the parent, which is either self (no parent) or the parent body
        # for Exec().
        return self._factory.make_body_context(self.context.body(), self.parent)


class WizardCancelContext(
    WizardInterpreterContext[ContextState, wizardParser.CancelContext]
):

    """
    Special context that is returned when a 'Cancel' instruction is encountered.
    """

    # The cancel message:
    _message: Optional[str]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._message = None
        if self.context.expr():
            self._message = self._factory.evisitor.visitExpr(
                self.context.expr(), self.state, str
            ).value

    def message(self) -> Optional[str]:
        """
        Returns:
            The cancel message, if any.
        """
        return self._message

    def exec_(self) -> WizardInterpreterContext:
        return self._factory.make_termination_context(self)


class WizardReturnContext(
    WizardInterpreterContext[ContextState, wizardParser.ReturnContext]
):

    """
    Special context that is returned when a 'Return' instruction is encountered.
    """

    def exec_(self) -> WizardInterpreterContext:
        return self._factory.make_termination_context(self)


class WizardBodyContext(
    WizardInterpreterContext[ContextState, wizardParser.BodyContext]
):

    """
    Body context.
    """

    # Index of the next child to process.
    _ichild: int

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._ichild = 0

    def exec_(self) -> WizardInterpreterContext:

        # Empty block or no more children:
        if not self._context.children or self._ichild == len(self._context.children):
            if isinstance(self._parent, WizardTopLevelContext):
                return self._factory.make_termination_context(self)
            return self._factory._copy_parent(self)

        child = self._context.children[self._ichild]

        # Copy the body and increment the child counter:
        body = self._factory._copy_context(self)
        body._ichild += 1

        # Expression - Standard context:
        if isinstance(child, wizardParser.ExprContext):
            # Specific handling of exec()
            if (
                isinstance(child, wizardParser.FunctionCallContext)
                and child.Identifier().getText() == "Exec"
            ):
                return self._factory.make_exec_context(child, body)

            self._factory.evisitor.visitExpr(child, body.state)
            return body

        if isinstance(child, wizardParser.StmtContext):
            if child.assignment():
                return self._factory.make_assignment_context(child.assignment(), body)
            elif child.compoundAssignment():
                return self._factory.make_compound_assignment_context(
                    child.compoundAssignment(), body
                )
            elif child.controlFlowStmt():
                return self._make_control_flow_context(child.controlFlowStmt(), body)
            elif child.keywordStmt():
                return self._factory.make_keyword_context(child.keywordStmt(), body)

        raise WizardParseError(child, f"Unknown context in body: {child}.")

    def _make_control_flow_context(
        self, ctx: wizardParser.ControlFlowStmtContext, parent: "WizardBodyContext"
    ) -> WizardInterpreterContext:

        if isinstance(ctx, wizardParser.ForContext):
            return self._factory.make_for_loop_context(ctx.forStmt(), parent)
        elif isinstance(ctx, wizardParser.WhileContext):
            return self._factory.make_while_loop_context(ctx.whileStmt(), parent)
        elif isinstance(ctx, wizardParser.IfContext):
            return self._factory.make_if_context(ctx.ifStmt(), parent)
        elif isinstance(ctx, wizardParser.SelectContext):
            return self._factory.make_select_context(ctx.selectStmt(), parent)
        elif isinstance(ctx, wizardParser.BreakContext):
            return self._factory.make_break_context(ctx, parent)
        elif isinstance(ctx, wizardParser.ContinueContext):
            return self._factory.make_continue_context(ctx, parent)
        elif isinstance(ctx, wizardParser.CancelContext):
            return self._factory.make_cancel_context(ctx.cancelStmt(), parent)
        elif isinstance(ctx, wizardParser.ReturnContext):
            return self._factory.make_return_context(ctx, parent)

        raise WizardParseError(ctx, f"Unknown control flow statement: {ctx}.")


class WizardBreakContext(
    WizardInterpreterContext[ContextState, wizardParser.BreakContext]
):
    def exec_(self) -> WizardInterpreterContext:

        # Find the first loop or case:
        loop_or_case: Optional[WizardInterpreterContext] = self

        while loop_or_case is not None and not isinstance(
            loop_or_case, WizardBreakableContext
        ):
            loop_or_case = loop_or_case.parent

        if loop_or_case is None:
            raise WizardParseError(
                self.context, "Invalid 'Break' statement encountered."
            )

        return self._factory._copy_context(loop_or_case.break_(), self.state)


class WizardContinueContext(
    WizardInterpreterContext[ContextState, wizardParser.ContinueContext]
):
    def exec_(self) -> WizardInterpreterContext:

        # Find the first parent loop:
        loop: Optional[WizardInterpreterContext] = self

        while loop is not None and not isinstance(loop, WizardContinuableContext):
            loop = loop.parent

        if loop is None:
            raise WizardParseError(
                self.context, "Invalid 'Continue' statement encountered."
            )

        return self._factory._copy_context(loop.continue_(), self.state)


class WizardExecContext(
    WizardInterpreterContext[ContextState, wizardParser.FunctionCallContext]
):
    def exec_(self) -> WizardInterpreterContext:
        # Import here otherwise we have circular imports:
        from ..utils import make_parse_wizard_context

        state = self.state.copy()

        # Parse the expression:
        args: List[wizardParser.ExprContext] = list(self.context.argList().expr())
        if len(args) != 1:
            raise WizardTypeError(
                self.context, f"Exec() expect exactly one argument, found {len(args)}."
            )
        script = self._factory.evisitor.visitExpr(args[0], state, str).value

        try:
            context = make_parse_wizard_context(script)
        except WizardError as we:
            # If an error occurs, the WizardError does not have a context, set it:
            we._ctx = self.context
            raise we

        return WizardTopLevelContext(self._factory, context, state, self.parent)


class WizardIfContext(
    WizardInterpreterContext[ContextState, wizardParser.IfStmtContext]
):
    def exec_(self) -> WizardInterpreterContext:

        parent = self._factory._copy_parent(self)

        if self._factory.evisitor.visitExpr(self.context.expr(), parent.state):
            return self._factory.make_body_context(self.context.body(), parent)
        else:
            for eifc in self.context.elifStmt():
                if self._factory.evisitor.visitExpr(eifc.expr(), parent.state):
                    return self._factory.make_body_context(eifc.body(), parent)

            if self.context.elseStmt():
                return self._factory.make_body_context(
                    self.context.elseStmt().body(),
                    parent,
                )

        return parent


class WizardForLoopContext(
    WizardInterpreterContext[ContextState, wizardParser.ForStmtContext],
    WizardBreakableContext,
    WizardContinuableContext,
):

    """
    For-loop context.
    """

    # Name of the loop variable, sequence of values and current index:
    _name: str
    _values: Sequence[Value]
    _index: int

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.context.forInHeader():
            self._values = self._for_in_header(self.context.forInHeader())
        else:
            self._values = self._for_range_header(self.context.forRangeHeader())

        self._name = self.context.Identifier().getText()
        self._index = 0

    def break_(self) -> WizardInterpreterContext:
        return self._factory._copy_parent(self)

    def continue_(self) -> WizardInterpreterContext:
        return self

    def _for_range_header(
        self, ctx: wizardParser.ForRangeHeaderContext
    ) -> Sequence[Value]:

        sta: Value[int] = self._factory.evisitor.visitExpr(ctx.expr(0), self.state, int)
        end: Value[int] = self._factory.evisitor.visitExpr(ctx.expr(1), self.state, int)

        if ctx.expr(2):
            by = self._factory.evisitor.visitExpr(ctx.expr(2), self.state, int)
        else:
            by = Value(1)

        return [Value(i) for i in range(sta.value, end.value + 1, by.value)]

    def _for_in_header(self, ctx: wizardParser.ForInHeaderContext) -> Sequence[Value]:

        value = self._factory.evisitor.visitExpr(ctx.expr(), self.state)

        if isinstance(value.value, SubPackage):
            return [Value(f) for f in value.value.files]
        elif isinstance(value.value, SubPackages):
            return [Value(sp) for sp in value.value]
        elif isinstance(value.value, str):
            # TODO: Allow?
            return [Value(s) for s in value.value]
        else:
            raise WizardTypeError(
                ctx, f"Cannot iterable over value of type {value.type}."
            )

    def exec_(self) -> WizardInterpreterContext:

        # End of the loop:
        if self._index == len(self._values):
            return self._factory._copy_parent(self)

        # Retrieve the next value and set it:
        value = self._values[self._index]

        # Copy the context:
        loop = self._factory._copy_context(self)
        loop._index += 1

        loop.state._variables[self._name] = value

        # Return a body context:
        return self._factory.make_body_context(self.context.body(), loop)


class WizardWhileLoopContext(
    WizardInterpreterContext[ContextState, wizardParser.WhileStmtContext],
    WizardBreakableContext,
    WizardContinuableContext,
):
    def break_(self) -> WizardInterpreterContext:
        return self._factory._copy_parent(self)

    def continue_(self) -> WizardInterpreterContext:
        return self

    def exec_(self) -> WizardInterpreterContext:

        loop = self._factory._copy_context(self)

        # If the expression evaluates to True, we return a body context:
        if self._factory.evisitor.visitExpr(self.context.expr(), loop.state):
            return self._factory.make_body_context(self.context.body(), loop)

        # Otherwise we return the parent (after updating the variables):
        return self._factory._copy_parent(loop)


class WizardAssignmentContext(
    WizardInterpreterContext[
        ContextState,
        wizardParser.AssignmentContext,
    ]
):
    def exec_(self) -> WizardInterpreterContext:

        parent = self._factory._copy_parent(self)

        # Retrieve the name:
        name: str = self.context.Identifier().getText()

        # Evaluate the expression:
        value: Value = self._factory.evisitor.visitExpr(
            self.context.expr(), parent.state
        )

        parent.state._variables[name] = value

        return parent


class WizardCompoundAssignmentContext(
    WizardInterpreterContext[
        ContextState,
        wizardParser.CompoundAssignmentContext,
    ]
):
    def exec_(self) -> WizardInterpreterContext:

        parent = self._factory._copy_parent(self)

        # Retrieve the name:
        name: str = self.context.Identifier().getText()

        # Evaluate the expression:
        value: Value = self._factory.evisitor.visitExpr(
            self.context.expr(), parent.state
        )

        if name not in parent.state._variables:
            raise WizardNameError(self.context.Identifier(), name)

        op: Callable[[Value, Value], Value]
        if self.context.CompoundExp():
            op = Value.__pow__
        elif self.context.CompoundMul():
            op = Value.__mul__
        elif self.context.CompoundDiv():
            op = Value.__div__
        elif self.context.CompoundMod():
            op = Value.__mod__
        elif self.context.CompoundAdd():
            op = Value.__add__
        elif self.context.CompoundSub():
            op = Value.__sub__
        else:
            raise WizardParseError(
                self.context, f"Unknown compound operation: {self.context}."
            )

        try:
            value = op(parent.state._variables[name], value)
        except TypeError as te:
            raise WizardTypeError(self.context, str(te))

        parent.state._variables[name] = value

        return parent


def parse_select_options(
    factory: "WizardInterpreterContextFactory",
    context: Union[wizardParser.SelectOneContext, wizardParser.SelectManyContext],
    state: ContextState,
) -> Tuple[str, List[SelectOption], List[SelectOption]]:

    # Parse the description and option:
    description = factory.evisitor.visitExpr(context.expr(), state, str).value

    options: List[SelectOption] = []
    defaults: List[SelectOption] = []
    for opt in context.optionTuple():
        a, b, c = (
            factory.evisitor.visitExpr(opt.expr(0), state, str),
            factory.evisitor.visitExpr(opt.expr(1), state, str),
            factory.evisitor.visitExpr(opt.expr(2), state, str),
        )

        name = a.value
        isdef = False
        if name.startswith("|"):
            name = name[1:]
            isdef = True

        options.append(
            SelectOption(name, b.value, c.value if c.value.strip() else None)
        )

        # Add to defaults:
        if isdef:
            defaults.append(options[-1])

    # Not default in SelectOne -> Select first one.
    if not isinstance(context, wizardParser.SelectManyContext) and not defaults:
        defaults.append(options[0])

    return description, options, defaults


class WizardSelectContext(WizardInterpreterContext[ContextState, RuleContext]):

    # The description:
    _description: str

    # The list of options:
    _options: List[SelectOption]

    # The default option(s):
    _default: List[SelectOption]

    # The selected option(s):
    _selected: List[SelectOption]

    def __init__(
        self,
        factory: "WizardInterpreterContextFactory",
        context: Union[wizardParser.SelectOneContext, wizardParser.SelectManyContext],
        parent: WizardInterpreterContext,
    ):
        super().__init__(factory, context, parent)

        self._description, self._options, self._defaults = parse_select_options(
            factory, context, self.state
        )

        # Note: It's the job of the child class to choose the default selected.

    @property
    def description(self) -> str:
        """
        Returns:
            The description for this select context.
        """
        return self._description

    @property
    def options(self) -> List[SelectOption]:
        """
        Returns:
            The available options for this select context.
        """
        return self._options

    def _select(self, options: List[SelectOption]) -> "WizardSelectContext":
        """
        Select the given options and return the object.

        Args:
            options: The options to select.

        Returns:
            A copy of the current object with the given options selected.
        """
        copy = self._factory._copy_context(self)
        copy._selected = options
        return copy

    def exec_(self) -> "WizardSelectCasesContext":
        # This completely delegates to the other context:
        return self._factory.make_select_cases_context(
            self._selected, self.context, self.parent
        )


class WizardSelectOneContext(
    WizardSelectContext[ContextState, wizardParser.SelectOneContext]
):
    def __init__(
        self,
        factory: "WizardInterpreterContextFactory",
        context: wizardParser.SelectOneContext,
        parent: WizardInterpreterContext,
    ):
        super().__init__(factory, context, parent)

        # More than one default in SelectOne:
        if len(self._defaults) != 1:
            self._factory.severity.raise_or_warn(
                Issue.MULTIPLE_DEFAULTS_IN_SELECT_ON,
                WizardTypeError(
                    context,
                    "SelectOne statement should have a single default value.",
                ),
                "SelectOne statement should have a single default, using the"
                " first one.",
            )
            self._defaults = [self._defaults[0]]

        # We select the defaults:
        self._selected = self._defaults

    @property
    def default(self) -> SelectOption:
        """
        Returns:
            The default option for this select context.
        """
        return self._defaults[0]

    def select(self, option: SelectOption) -> "WizardSelectOneContext":
        """
        Select the given option and return the object.

        Args:
            option: The option to select.

        Returns:
            A copy of the current object with the given option selected.
        """
        return self._select([option])  # type: ignore


class WizardSelectManyContext(
    WizardSelectContext[ContextState, wizardParser.SelectOneContext]
):
    def __init__(
        self,
        factory: "WizardInterpreterContextFactory",
        context: wizardParser.SelectManyContext,
        parent: WizardInterpreterContext,
    ):
        super().__init__(factory, context, parent)

        # We select the defaults:
        self._selected = self._defaults

    @property
    def defaults(self) -> List[SelectOption]:
        """
        Returns:
            The default options for this select context.
        """
        return self._defaults

    def select(self, options: List[SelectOption]) -> "WizardSelectManyContext":
        """
        Select the given list of options and return the object.

        Args:
            options: The options to select.

        Returns:
            A copy of the current object with the given options selected.
        """
        return self._select(options)  # type: ignore


class WizardSelectCasesContext(
    WizardInterpreterContext[ContextState, wizardParser.SelectStmtContext],
    WizardBreakableContext,
):

    # The list of selected options:
    _options: List[SelectOption]

    # The list of cases, the current index, a boolean indicating if
    # the value has been found, and a boolean for fallthrough:
    _ismany: bool
    _cases: List[wizardParser.CaseStmtContext]
    _default: Optional[wizardParser.DefaultStmtContext]
    _index: int
    _found: bool
    _fallthrough: bool

    def __init__(
        self,
        factory: "WizardInterpreterContextFactory",
        options: List[SelectOption],
        context: Union[wizardParser.SelectOneContext, wizardParser.SelectManyContext],
        parent: WizardInterpreterContext,
    ):
        """
        Args:
            factory: The context factory.
            options: List of selected options.
            context: The select context.
            parent: The parent context.
        """
        super().__init__(factory, context, parent)

        self._options = options

        if isinstance(self.context, wizardParser.SelectOneContext):
            self._ismany = False
        else:
            self._ismany = True

        self._cases = list(context.selectCaseList().caseStmt())
        self._default = context.selectCaseList().defaultStmt()

        self._index = 0
        self._found = False
        self._fallthrough = False

    def break_(self) -> WizardInterpreterContext:
        # Disable fallthrough:
        copy = self._factory._copy_context(self)
        copy._fallthrough = False

        # We return the context itself to keep evaluating cases:
        return copy

    def exec_(self) -> WizardInterpreterContext:

        # Element found in a selectOne, returns to the parent:
        if self._found and not self._ismany and not self._fallthrough:
            return self._factory._copy_parent(self)

        # Copy the current context:
        snext = self._factory._copy_context(self)

        # Cases remaining:
        if self._index < len(self._cases):

            # Find the next case:
            case = self._cases[self._index]
            snext._index += 1

            target: Value[str] = self._factory.evisitor.visitExpr(
                case.expr(), snext.state, str
            )

            # Check if the case match or if we have a fallthrough:
            if self._fallthrough or any(
                sopt.name == target.value for sopt in self._options
            ):

                # We found the value, remember it:
                snext._found = True

                # Assume fallthrough, the break_() will set this off if a 'Break'
                # statement is found:
                snext._fallthrough = True

                return self._factory.make_case_context(case, snext)

            # Return the context itself:
            return snext

        # No cases remaining, default, and not found:
        elif self._default and (self._fallthrough or not self._found):
            snext._found = True
            return self._factory.make_case_context(self._default, snext)

        # No need to copy from snext since if we reach this return statement, nothing
        # has been updated in snext.
        return self._factory._copy_parent(self)


class WizardCaseContext(
    WizardInterpreterContext[
        ContextState,
        Union[wizardParser.CaseStmtContext, wizardParser.DefaultStmtContext],
    ]
):
    def exec_(self) -> WizardInterpreterContext:
        # It's the select context job to check if this should be executed:
        return self._factory.make_body_context(self.context.body(), self.parent)



# Generated from ./vendor/wizparse/wizards/wizard.g4 by ANTLR 4.10.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .wizardParser import wizardParser
else:
    from wizardParser import wizardParser

# This class defines a complete listener for a parse tree produced by wizardParser.
class wizardListener(ParseTreeListener):

    # Enter a parse tree produced by wizardParser#parseWizard.
    def enterParseWizard(self, ctx:wizardParser.ParseWizardContext):
        pass

    # Exit a parse tree produced by wizardParser#parseWizard.
    def exitParseWizard(self, ctx:wizardParser.ParseWizardContext):
        pass


    # Enter a parse tree produced by wizardParser#body.
    def enterBody(self, ctx:wizardParser.BodyContext):
        pass

    # Exit a parse tree produced by wizardParser#body.
    def exitBody(self, ctx:wizardParser.BodyContext):
        pass


    # Enter a parse tree produced by wizardParser#stmt.
    def enterStmt(self, ctx:wizardParser.StmtContext):
        pass

    # Exit a parse tree produced by wizardParser#stmt.
    def exitStmt(self, ctx:wizardParser.StmtContext):
        pass


    # Enter a parse tree produced by wizardParser#assignment.
    def enterAssignment(self, ctx:wizardParser.AssignmentContext):
        pass

    # Exit a parse tree produced by wizardParser#assignment.
    def exitAssignment(self, ctx:wizardParser.AssignmentContext):
        pass


    # Enter a parse tree produced by wizardParser#compoundAssignment.
    def enterCompoundAssignment(self, ctx:wizardParser.CompoundAssignmentContext):
        pass

    # Exit a parse tree produced by wizardParser#compoundAssignment.
    def exitCompoundAssignment(self, ctx:wizardParser.CompoundAssignmentContext):
        pass


    # Enter a parse tree produced by wizardParser#Break.
    def enterBreak(self, ctx:wizardParser.BreakContext):
        pass

    # Exit a parse tree produced by wizardParser#Break.
    def exitBreak(self, ctx:wizardParser.BreakContext):
        pass


    # Enter a parse tree produced by wizardParser#Cancel.
    def enterCancel(self, ctx:wizardParser.CancelContext):
        pass

    # Exit a parse tree produced by wizardParser#Cancel.
    def exitCancel(self, ctx:wizardParser.CancelContext):
        pass


    # Enter a parse tree produced by wizardParser#Continue.
    def enterContinue(self, ctx:wizardParser.ContinueContext):
        pass

    # Exit a parse tree produced by wizardParser#Continue.
    def exitContinue(self, ctx:wizardParser.ContinueContext):
        pass


    # Enter a parse tree produced by wizardParser#For.
    def enterFor(self, ctx:wizardParser.ForContext):
        pass

    # Exit a parse tree produced by wizardParser#For.
    def exitFor(self, ctx:wizardParser.ForContext):
        pass


    # Enter a parse tree produced by wizardParser#If.
    def enterIf(self, ctx:wizardParser.IfContext):
        pass

    # Exit a parse tree produced by wizardParser#If.
    def exitIf(self, ctx:wizardParser.IfContext):
        pass


    # Enter a parse tree produced by wizardParser#Return.
    def enterReturn(self, ctx:wizardParser.ReturnContext):
        pass

    # Exit a parse tree produced by wizardParser#Return.
    def exitReturn(self, ctx:wizardParser.ReturnContext):
        pass


    # Enter a parse tree produced by wizardParser#Select.
    def enterSelect(self, ctx:wizardParser.SelectContext):
        pass

    # Exit a parse tree produced by wizardParser#Select.
    def exitSelect(self, ctx:wizardParser.SelectContext):
        pass


    # Enter a parse tree produced by wizardParser#While.
    def enterWhile(self, ctx:wizardParser.WhileContext):
        pass

    # Exit a parse tree produced by wizardParser#While.
    def exitWhile(self, ctx:wizardParser.WhileContext):
        pass


    # Enter a parse tree produced by wizardParser#cancelStmt.
    def enterCancelStmt(self, ctx:wizardParser.CancelStmtContext):
        pass

    # Exit a parse tree produced by wizardParser#cancelStmt.
    def exitCancelStmt(self, ctx:wizardParser.CancelStmtContext):
        pass


    # Enter a parse tree produced by wizardParser#caseStmt.
    def enterCaseStmt(self, ctx:wizardParser.CaseStmtContext):
        pass

    # Exit a parse tree produced by wizardParser#caseStmt.
    def exitCaseStmt(self, ctx:wizardParser.CaseStmtContext):
        pass


    # Enter a parse tree produced by wizardParser#defaultStmt.
    def enterDefaultStmt(self, ctx:wizardParser.DefaultStmtContext):
        pass

    # Exit a parse tree produced by wizardParser#defaultStmt.
    def exitDefaultStmt(self, ctx:wizardParser.DefaultStmtContext):
        pass


    # Enter a parse tree produced by wizardParser#elifStmt.
    def enterElifStmt(self, ctx:wizardParser.ElifStmtContext):
        pass

    # Exit a parse tree produced by wizardParser#elifStmt.
    def exitElifStmt(self, ctx:wizardParser.ElifStmtContext):
        pass


    # Enter a parse tree produced by wizardParser#elseStmt.
    def enterElseStmt(self, ctx:wizardParser.ElseStmtContext):
        pass

    # Exit a parse tree produced by wizardParser#elseStmt.
    def exitElseStmt(self, ctx:wizardParser.ElseStmtContext):
        pass


    # Enter a parse tree produced by wizardParser#forStmt.
    def enterForStmt(self, ctx:wizardParser.ForStmtContext):
        pass

    # Exit a parse tree produced by wizardParser#forStmt.
    def exitForStmt(self, ctx:wizardParser.ForStmtContext):
        pass


    # Enter a parse tree produced by wizardParser#forRangeHeader.
    def enterForRangeHeader(self, ctx:wizardParser.ForRangeHeaderContext):
        pass

    # Exit a parse tree produced by wizardParser#forRangeHeader.
    def exitForRangeHeader(self, ctx:wizardParser.ForRangeHeaderContext):
        pass


    # Enter a parse tree produced by wizardParser#forInHeader.
    def enterForInHeader(self, ctx:wizardParser.ForInHeaderContext):
        pass

    # Exit a parse tree produced by wizardParser#forInHeader.
    def exitForInHeader(self, ctx:wizardParser.ForInHeaderContext):
        pass


    # Enter a parse tree produced by wizardParser#ifStmt.
    def enterIfStmt(self, ctx:wizardParser.IfStmtContext):
        pass

    # Exit a parse tree produced by wizardParser#ifStmt.
    def exitIfStmt(self, ctx:wizardParser.IfStmtContext):
        pass


    # Enter a parse tree produced by wizardParser#selectStmt.
    def enterSelectStmt(self, ctx:wizardParser.SelectStmtContext):
        pass

    # Exit a parse tree produced by wizardParser#selectStmt.
    def exitSelectStmt(self, ctx:wizardParser.SelectStmtContext):
        pass


    # Enter a parse tree produced by wizardParser#selectCaseList.
    def enterSelectCaseList(self, ctx:wizardParser.SelectCaseListContext):
        pass

    # Exit a parse tree produced by wizardParser#selectCaseList.
    def exitSelectCaseList(self, ctx:wizardParser.SelectCaseListContext):
        pass


    # Enter a parse tree produced by wizardParser#optionTuple.
    def enterOptionTuple(self, ctx:wizardParser.OptionTupleContext):
        pass

    # Exit a parse tree produced by wizardParser#optionTuple.
    def exitOptionTuple(self, ctx:wizardParser.OptionTupleContext):
        pass


    # Enter a parse tree produced by wizardParser#selectOne.
    def enterSelectOne(self, ctx:wizardParser.SelectOneContext):
        pass

    # Exit a parse tree produced by wizardParser#selectOne.
    def exitSelectOne(self, ctx:wizardParser.SelectOneContext):
        pass


    # Enter a parse tree produced by wizardParser#selectMany.
    def enterSelectMany(self, ctx:wizardParser.SelectManyContext):
        pass

    # Exit a parse tree produced by wizardParser#selectMany.
    def exitSelectMany(self, ctx:wizardParser.SelectManyContext):
        pass


    # Enter a parse tree produced by wizardParser#whileStmt.
    def enterWhileStmt(self, ctx:wizardParser.WhileStmtContext):
        pass

    # Exit a parse tree produced by wizardParser#whileStmt.
    def exitWhileStmt(self, ctx:wizardParser.WhileStmtContext):
        pass


    # Enter a parse tree produced by wizardParser#argList.
    def enterArgList(self, ctx:wizardParser.ArgListContext):
        pass

    # Exit a parse tree produced by wizardParser#argList.
    def exitArgList(self, ctx:wizardParser.ArgListContext):
        pass


    # Enter a parse tree produced by wizardParser#keywordStmt.
    def enterKeywordStmt(self, ctx:wizardParser.KeywordStmtContext):
        pass

    # Exit a parse tree produced by wizardParser#keywordStmt.
    def exitKeywordStmt(self, ctx:wizardParser.KeywordStmtContext):
        pass


    # Enter a parse tree produced by wizardParser#PreIncrement.
    def enterPreIncrement(self, ctx:wizardParser.PreIncrementContext):
        pass

    # Exit a parse tree produced by wizardParser#PreIncrement.
    def exitPreIncrement(self, ctx:wizardParser.PreIncrementContext):
        pass


    # Enter a parse tree produced by wizardParser#DotFunctionCall.
    def enterDotFunctionCall(self, ctx:wizardParser.DotFunctionCallContext):
        pass

    # Exit a parse tree produced by wizardParser#DotFunctionCall.
    def exitDotFunctionCall(self, ctx:wizardParser.DotFunctionCallContext):
        pass


    # Enter a parse tree produced by wizardParser#Or.
    def enterOr(self, ctx:wizardParser.OrContext):
        pass

    # Exit a parse tree produced by wizardParser#Or.
    def exitOr(self, ctx:wizardParser.OrContext):
        pass


    # Enter a parse tree produced by wizardParser#In.
    def enterIn(self, ctx:wizardParser.InContext):
        pass

    # Exit a parse tree produced by wizardParser#In.
    def exitIn(self, ctx:wizardParser.InContext):
        pass


    # Enter a parse tree produced by wizardParser#PostDecrement.
    def enterPostDecrement(self, ctx:wizardParser.PostDecrementContext):
        pass

    # Exit a parse tree produced by wizardParser#PostDecrement.
    def exitPostDecrement(self, ctx:wizardParser.PostDecrementContext):
        pass


    # Enter a parse tree produced by wizardParser#TimesDivideModulo.
    def enterTimesDivideModulo(self, ctx:wizardParser.TimesDivideModuloContext):
        pass

    # Exit a parse tree produced by wizardParser#TimesDivideModulo.
    def exitTimesDivideModulo(self, ctx:wizardParser.TimesDivideModuloContext):
        pass


    # Enter a parse tree produced by wizardParser#Index.
    def enterIndex(self, ctx:wizardParser.IndexContext):
        pass

    # Exit a parse tree produced by wizardParser#Index.
    def exitIndex(self, ctx:wizardParser.IndexContext):
        pass


    # Enter a parse tree produced by wizardParser#Exponentiation.
    def enterExponentiation(self, ctx:wizardParser.ExponentiationContext):
        pass

    # Exit a parse tree produced by wizardParser#Exponentiation.
    def exitExponentiation(self, ctx:wizardParser.ExponentiationContext):
        pass


    # Enter a parse tree produced by wizardParser#PlusMinus.
    def enterPlusMinus(self, ctx:wizardParser.PlusMinusContext):
        pass

    # Exit a parse tree produced by wizardParser#PlusMinus.
    def exitPlusMinus(self, ctx:wizardParser.PlusMinusContext):
        pass


    # Enter a parse tree produced by wizardParser#Not.
    def enterNot(self, ctx:wizardParser.NotContext):
        pass

    # Exit a parse tree produced by wizardParser#Not.
    def exitNot(self, ctx:wizardParser.NotContext):
        pass


    # Enter a parse tree produced by wizardParser#Lesser.
    def enterLesser(self, ctx:wizardParser.LesserContext):
        pass

    # Exit a parse tree produced by wizardParser#Lesser.
    def exitLesser(self, ctx:wizardParser.LesserContext):
        pass


    # Enter a parse tree produced by wizardParser#Negative.
    def enterNegative(self, ctx:wizardParser.NegativeContext):
        pass

    # Exit a parse tree produced by wizardParser#Negative.
    def exitNegative(self, ctx:wizardParser.NegativeContext):
        pass


    # Enter a parse tree produced by wizardParser#Equal.
    def enterEqual(self, ctx:wizardParser.EqualContext):
        pass

    # Exit a parse tree produced by wizardParser#Equal.
    def exitEqual(self, ctx:wizardParser.EqualContext):
        pass


    # Enter a parse tree produced by wizardParser#And.
    def enterAnd(self, ctx:wizardParser.AndContext):
        pass

    # Exit a parse tree produced by wizardParser#And.
    def exitAnd(self, ctx:wizardParser.AndContext):
        pass


    # Enter a parse tree produced by wizardParser#Slice.
    def enterSlice(self, ctx:wizardParser.SliceContext):
        pass

    # Exit a parse tree produced by wizardParser#Slice.
    def exitSlice(self, ctx:wizardParser.SliceContext):
        pass


    # Enter a parse tree produced by wizardParser#Value.
    def enterValue(self, ctx:wizardParser.ValueContext):
        pass

    # Exit a parse tree produced by wizardParser#Value.
    def exitValue(self, ctx:wizardParser.ValueContext):
        pass


    # Enter a parse tree produced by wizardParser#Greater.
    def enterGreater(self, ctx:wizardParser.GreaterContext):
        pass

    # Exit a parse tree produced by wizardParser#Greater.
    def exitGreater(self, ctx:wizardParser.GreaterContext):
        pass


    # Enter a parse tree produced by wizardParser#FunctionCall.
    def enterFunctionCall(self, ctx:wizardParser.FunctionCallContext):
        pass

    # Exit a parse tree produced by wizardParser#FunctionCall.
    def exitFunctionCall(self, ctx:wizardParser.FunctionCallContext):
        pass


    # Enter a parse tree produced by wizardParser#PostIncrement.
    def enterPostIncrement(self, ctx:wizardParser.PostIncrementContext):
        pass

    # Exit a parse tree produced by wizardParser#PostIncrement.
    def exitPostIncrement(self, ctx:wizardParser.PostIncrementContext):
        pass


    # Enter a parse tree produced by wizardParser#ParenExpr.
    def enterParenExpr(self, ctx:wizardParser.ParenExprContext):
        pass

    # Exit a parse tree produced by wizardParser#ParenExpr.
    def exitParenExpr(self, ctx:wizardParser.ParenExprContext):
        pass


    # Enter a parse tree produced by wizardParser#PreDecrement.
    def enterPreDecrement(self, ctx:wizardParser.PreDecrementContext):
        pass

    # Exit a parse tree produced by wizardParser#PreDecrement.
    def exitPreDecrement(self, ctx:wizardParser.PreDecrementContext):
        pass


    # Enter a parse tree produced by wizardParser#constant.
    def enterConstant(self, ctx:wizardParser.ConstantContext):
        pass

    # Exit a parse tree produced by wizardParser#constant.
    def exitConstant(self, ctx:wizardParser.ConstantContext):
        pass


    # Enter a parse tree produced by wizardParser#integer.
    def enterInteger(self, ctx:wizardParser.IntegerContext):
        pass

    # Exit a parse tree produced by wizardParser#integer.
    def exitInteger(self, ctx:wizardParser.IntegerContext):
        pass


    # Enter a parse tree produced by wizardParser#decimal.
    def enterDecimal(self, ctx:wizardParser.DecimalContext):
        pass

    # Exit a parse tree produced by wizardParser#decimal.
    def exitDecimal(self, ctx:wizardParser.DecimalContext):
        pass


    # Enter a parse tree produced by wizardParser#string.
    def enterString(self, ctx:wizardParser.StringContext):
        pass

    # Exit a parse tree produced by wizardParser#string.
    def exitString(self, ctx:wizardParser.StringContext):
        pass



del wizardParser


# -*- encoding: utf-8 -*-

from enum import Enum
from typing import Generic, Optional

from .contexts.keywords import WizardKeywordContext
from .severity import SeverityContext
from .state import ContextState


class WizardKeyword(Enum):

    """
    Enumeration representing possible keywords in a Wizard scripts.
    """

    DESELECT_ALL = "DeSelectAll"
    DESELECT_ALL_PLUGINS = "DeSelectAllPlugins"
    DESELECT_PLUGIN = "DeSelectPlugin"
    DESELECT_SUBPACKAGE = "DeSelectSubPackage"
    NOTE = "Note"
    RENAME_PLUGIN = "RenamePlugin"
    REQUIRE_VERSIONS = "RequireVersions"
    RESET_ALL_PLUGIN_NAMES = "ResetAllPluginNames"
    RESET_PLUGIN_NAME = "ResetPluginName"
    SELECT_ALL = "SelectAll"
    SELECT_ALL_PLUGINS = "SelectAllPlugins"
    SELECT_PLUGIN = "SelectPlugin"
    SELECT_SUBPACKAGE = "SelectSubPackage"


class WizardKeywordVisitor(Generic[ContextState]):

    """
    Keyword visitor that can be extended to perform specific actions for
    each keywords. By default, none of the methods do anything.
    """

    _severity: SeverityContext

    def __init__(self, severity: SeverityContext):
        self._severity = severity

    def visitDeSelectAll(self, context: WizardKeywordContext, state: ContextState):
        """
        Args:
            context: The context corresponding to the keyword.
            state: The interpreter state to update.
        """
        ...

    def visitDeSelectAllPlugins(
        self, context: WizardKeywordContext, state: ContextState
    ):
        """
        Args:
            context: The context corresponding to the keyword.
            state: The interpreter state to update.
        """
        ...

    def visitDeSelectPlugin(
        self, context: WizardKeywordContext, state: ContextState, name: str
    ):
        """
        Args:
            context: The context corresponding to the keyword.
            state: The interpreter state to update.
            name: The name of the plugin to de-select.
        """
        ...

    def visitDeSelectSubPackage(
        self, context: WizardKeywordContext, state: ContextState, name: str
    ):
        """
        Args:
            context: The context corresponding to the keyword.
            state: The interpreter state to update.
            name: The name of the subpackage to de-select.
        """
        ...

    def visitNote(self, context: WizardKeywordContext, state: ContextState, text: str):
        """
        Args:
            context: The context corresponding to the keyword.
            state: The interpreter state to update.
            text: The text for the note.
        """
        ...

    def visitRenamePlugin(
        self,
        context: WizardKeywordContext,
        state: ContextState,
        original_name: str,
        new_name: str,
    ):
        """
        Args:
            context: The context corresponding to the keyword.
            state: The interpreter state to update.
            original_name: The original name of the plugin.
            new_name: The new name of the plugin.
        """
        ...

    def visitRequireVersions(
        self,
        state: ContextState,
        game_version: str,
        script_extender_version: Optional[str],
        graphics_extender_version: Optional[str],
        wrye_bash_version: Optional[str],
    ):
        """
        Args:
            context: The context corresponding to the keyword.
            state: The interpreter state to update.
            game_version: The required game version.
            script_extender_version: The required script extender version.
            graphics_extender_version: The required graphics extender version.
            wrye_bash_version: The required wrye bash version.
        """
        ...

    def visitResetAllPluginNames(
        self, context: WizardKeywordContext, state: ContextState
    ):
        """
        Args:
            context: The context corresponding to the keyword.
            state: The interpreter state to update.
        """
        ...

    def visitResetPluginName(
        self, context: WizardKeywordContext, state: ContextState, name: str
    ):
        """
        Args:
            context: The context corresponding to the keyword.
            state: The interpreter state to update.
            name: The original name of the plugin.
        """
        ...

    def visitSelectAll(self, context: WizardKeywordContext, state: ContextState):
        """
        Args:
            context: The context corresponding to the keyword.
            state: The interpreter state to update.
        """
        ...

    def visitSelectAllPlugins(self, context: WizardKeywordContext, state: ContextState):
        """
        Args:
            context: The context corresponding to the keyword.
            state: The interpreter state to update.
        """
        ...

    def visitSelectPlugin(
        self, context: WizardKeywordContext, state: ContextState, name: str
    ):
        """
        Args:
            context: The context corresponding to the keyword.
            state: The interpreter state to update.
            name: The name of the plugin to select.
        """
        ...

    def visitSelectSubPackage(
        self, context: WizardKeywordContext, state: ContextState, name: str
    ):
        """
        Args:
            context: The context corresponding to the keyword.
            state: The interpreter state to update.
            name: The name of the subpackage to select.
        """
        ...



# -*- encoding: utf-8 -*-

import codecs
import re
from typing import Callable, List, Mapping, Optional, Type, overload

from .antlr4.wizardParser import wizardParser
from .errors import WizardIndexError, WizardNameError, WizardParseError, WizardTypeError
from .severity import Issue, SeverityContext
from .state import WizardInterpreterState
from .value import (  # noqa: F401
    SubPackage,
    SubPackages,
    Value,
    ValueType,
    VariableType,
    Void,
)


class WizardExpressionVisitor:

    """
    Visitor for Wizard expression. This visitor can be used to interpret Wizard
    expression and return `Value` object. The main entry point is `visitExpr()`
    which takes an expression context and returns a `Value`.
    """

    BAD_ESCAPE_SEQUENCE = re.compile(r"(^|(?<=[^\\]))\\(?=[^abfnrtuUx0-7\\])")

    _subpackages: SubPackages
    _functions: Mapping[str, Callable[[WizardInterpreterState, List[Value]], Value]]
    _severity: SeverityContext

    def __init__(
        self,
        subpackages: SubPackages,
        functions: Mapping[str, Callable[[WizardInterpreterState, List[Value]], Value]],
        severity: SeverityContext,
    ):
        """
        Args:
            subpackages: The subpackages for the visitor.
            functions: The list of usable functions.
            severity: The severity context.
        """
        self._subpackages = subpackages
        self._functions = functions
        self._severity = severity

    def visitTimesDivideModulo(
        self,
        ctx: wizardParser.TimesDivideModuloContext,
        state: WizardInterpreterState,
    ) -> Value:
        op: Callable[[Value, Value], Value] = Value.__mul__
        if ctx.Divide():
            op = Value.__div__
        elif ctx.Modulo():
            op = Value.__mod__
        return op(
            self.visitExpr(ctx.expr(0), state),
            self.visitExpr(ctx.expr(1), state),
        )

    def visitPlusMinus(
        self, ctx: wizardParser.PlusMinusContext, state: WizardInterpreterState
    ) -> Value:
        op: Callable[[Value, Value], Value] = Value.__add__
        if ctx.Minus():
            op = Value.__sub__
        return op(
            self.visitExpr(ctx.expr(0), state),
            self.visitExpr(ctx.expr(1), state),
        )

    def visitOr(
        self, ctx: wizardParser.OrContext, state: WizardInterpreterState
    ) -> Value:
        return self.visitExpr(ctx.expr(0), state) | self.visitExpr(ctx.expr(1), state)

    def visitFunctionCall(
        self,
        ctx: wizardParser.FunctionCallContext,
        state: WizardInterpreterState,
    ) -> Value:

        name = ctx.Identifier().getText()

        # Specific handle for Exec:
        if name == "Exec":
            raise WizardNameError(ctx, "Exec() cannot be used in a complex expression.")
        if name == "EndExec":
            raise WizardNameError(ctx, "EndExec() should not be used explicitly.")

        function: Callable[[WizardInterpreterState, List[Value]], Value]
        if name in self._functions:
            function = self._functions[name]
            is_visit = False
        elif hasattr(self, "visit" + name):
            function = getattr(self, "visit" + name)
            is_visit = True
        else:
            raise WizardNameError(ctx, name)

        values: List[Value] = []
        if ctx.argList():
            for ex in ctx.argList().expr():
                values.append(self.visitExpr(ex, state))

        if is_visit:
            return Value(function(state, *(value.value for value in values)))
        return function(state, values)

    def visitDotFunctionCall(
        self,
        ctx: wizardParser.DotFunctionCallContext,
        state: WizardInterpreterState,
    ) -> Value:
        values: List[Value] = [self.visitExpr(ctx.expr(), state)]

        mname = "{}.{}".format(values[0].type, ctx.Identifier().getText())
        if mname not in self._functions:
            raise WizardNameError(ctx, mname)

        if ctx.argList():
            for ex in ctx.argList().expr():
                values.append(self.visitExpr(ex, state))

        return self._functions[mname](state, values)

    def visitIn(
        self, ctx: wizardParser.InContext, state: WizardInterpreterState
    ) -> Value:
        return self.visitExpr(ctx.expr(1), state).contains(
            self.visitExpr(ctx.expr(0), state), bool(ctx.Colon())
        )

    def doCmp(
        self,
        op: Callable[[Value, Value], Value],
        lhs: Value,
        rhs: Value,
        case_insensitive: bool = True,
    ) -> Value:
        if case_insensitive:
            if not isinstance(lhs.value, str) or not isinstance(rhs.value, str):
                raise WizardTypeError(
                    "Cannot use case-insensitive comparison with values of type"
                    f" {lhs.type}, {rhs.type}."
                )
            lhs = Value(lhs.value.lower())
            rhs = Value(rhs.value.lower())
        return op(lhs, rhs)

    def visitEqual(
        self, ctx: wizardParser.EqualContext, state: WizardInterpreterState
    ) -> Value:
        op = Value.equals
        if ctx.NotEqual():
            op = Value.not_equals
        return self.doCmp(
            op,
            self.visitExpr(ctx.expr(0), state),
            self.visitExpr(ctx.expr(1), state),
            bool(ctx.Colon()),
        )

    def visitLesser(
        self, ctx: wizardParser.LesserContext, state: WizardInterpreterState
    ) -> Value:
        op = Value.__le__
        if ctx.Lesser():
            op = Value.__lt__
        return self.doCmp(
            op,
            self.visitExpr(ctx.expr(0), state),
            self.visitExpr(ctx.expr(1), state),
            bool(ctx.Colon()),
        )

    def visitGreater(
        self, ctx: wizardParser.GreaterContext, state: WizardInterpreterState
    ) -> Value:
        op = Value.__ge__
        if ctx.Greater():
            op = Value.__gt__
        return self.doCmp(
            op,
            self.visitExpr(ctx.expr(0), state),
            self.visitExpr(ctx.expr(1), state),
            bool(ctx.Colon()),
        )

    def visitExponentiation(
        self,
        ctx: wizardParser.ExponentiationContext,
        state: WizardInterpreterState,
    ) -> Value:
        return self.visitExpr(ctx.expr(0), state) ** self.visitExpr(ctx.expr(1), state)

    def visitIndex(
        self, ctx: wizardParser.IndexContext, state: WizardInterpreterState
    ) -> Value:
        return self.visitExpr(ctx.expr(0), state)[self.visitExpr(ctx.expr(1), state)]

    def visitPreDecrement(
        self,
        ctx: wizardParser.PreDecrementContext,
        state: WizardInterpreterState,
    ) -> Value:
        name = ctx.Identifier().getText()
        if name not in state._variables:
            raise WizardNameError(ctx, name)
        state._variables[ctx.Identifier().getText()] -= Value(1)
        return state._variables[ctx.Identifier().getText()]

    def visitPreIncrement(
        self,
        ctx: wizardParser.PreIncrementContext,
        state: WizardInterpreterState,
    ) -> Value:
        name = ctx.Identifier().getText()
        if name not in state._variables:
            raise WizardNameError(ctx, name)
        state._variables[ctx.Identifier().getText()] += Value(1)
        return state._variables[ctx.Identifier().getText()]

    def visitPostIncrement(
        self,
        ctx: wizardParser.PostIncrementContext,
        state: WizardInterpreterState,
    ) -> Value:
        name = ctx.Identifier().getText()
        if name not in state._variables:
            raise WizardNameError(ctx, name)
        state._variables[ctx.Identifier().getText()] += Value(1)
        return state._variables[ctx.Identifier().getText()]

    def visitPostDecrement(
        self,
        ctx: wizardParser.PostDecrementContext,
        state: WizardInterpreterState,
    ) -> Value:
        name = ctx.Identifier().getText()
        if name not in state._variables:
            raise WizardNameError(ctx, name)
        state._variables[ctx.Identifier().getText()] -= Value(1)
        return state._variables[ctx.Identifier().getText()]

    def visitNegative(
        self, ctx: wizardParser.NegativeContext, state: WizardInterpreterState
    ) -> Value:
        return -self.visitExpr(ctx.expr(), state)

    def visitNot(
        self, ctx: wizardParser.NotContext, state: WizardInterpreterState
    ) -> Value:
        return self.visitExpr(ctx.expr(), state).logical_not()

    def visitParenExpr(
        self, ctx: wizardParser.ParenExprContext, state: WizardInterpreterState
    ) -> Value:
        return self.visitExpr(ctx.expr(), state)

    def visitSlice(
        self, ctx: wizardParser.SliceContext, state: WizardInterpreterState
    ) -> Value:
        # expr LeftBracket expr? Colon expr? (Colon expr?)? RightBracket
        lhs = self.visitExpr(ctx.expr(0), state)

        start: Optional[Value] = None
        end: Optional[Value] = None
        step: Optional[Value] = None

        # Index of end expression in ctx.children (without start by
        # default):
        cidx: int = 3

        # Is there a start?
        if isinstance(ctx.children[2], wizardParser.ExprContext):
            start = self.visitExpr(ctx.expr(1), state)
            cidx = 4

        # Is there and end?
        if isinstance(ctx.children[cidx], wizardParser.ExprContext):
            end = self.visitExpr(ctx.children[cidx], state)

        # Is there a step?
        if ctx.Colon(1) and isinstance(ctx.children[-2], wizardParser.ExprContext):
            step = self.visitExpr(ctx.children[-2], state)

        # We keep None and let python built-in slice() do the job for us.
        return lhs.slice(start, end, step)

    def visitAnd(
        self, ctx: wizardParser.AndContext, state: WizardInterpreterState
    ) -> Value:
        lhs, rhs = self.visitExpr(ctx.expr(0), state), self.visitExpr(
            ctx.expr(1), state
        )
        return lhs & rhs

    def visitValue(
        self, ctx: wizardParser.ValueContext, state: WizardInterpreterState
    ) -> Value:
        if ctx.constant():
            return self.visitConstant(ctx.constant())
        if ctx.integer():
            return self.visitInteger(ctx.integer())
        if ctx.decimal():
            return self.visitDecimal(ctx.decimal())
        if ctx.string():
            return self.visitString(ctx.string())
        if ctx.Identifier():
            name = ctx.Identifier().getText()
            if name not in state._variables:
                # Severity check:
                self._severity.raise_or_warn(
                    Issue.USAGE_OF_NOTSET_VARIABLES,
                    WizardNameError(ctx, name),
                    f"Variable {name} used before being set, default to 0.",
                )
                return Value(0)
            else:
                return state._variables[ctx.Identifier().getText()]

        raise WizardNameError(ctx, ctx.getText())

    def visitConstant(self, ctx: wizardParser.ConstantContext) -> Value:
        if ctx.getText() == "False":
            return Value(False)
        elif ctx.getText() == "True":
            return Value(True)
        else:
            return Value(self._subpackages)

    def visitInteger(self, ctx: wizardParser.IntegerContext) -> Value:
        return Value(int(ctx.getText()))

    def visitDecimal(self, ctx: wizardParser.DecimalContext) -> Value:
        return Value(float(ctx.getText()))

    def visitString(self, ctx: wizardParser.StringContext) -> Value:
        # Remove the quotation marks:
        txt = ctx.getText()[1:-1]

        # Remove bad escape sequences:
        txt = self.BAD_ESCAPE_SEQUENCE.sub("", txt)

        # Replace escape sequences by Python escape sequences:
        txt = codecs.decode(txt, "unicode_escape")

        return Value(txt)

    @overload
    def visitExpr(
        self, ctx: wizardParser.ExprContext, state: WizardInterpreterState
    ) -> Value:
        ...

    @overload
    def visitExpr(
        self,
        ctx: wizardParser.ExprContext,
        state: WizardInterpreterState,
        typ: Type[ValueType],
    ) -> Value[ValueType]:
        ...

    def visitExpr(
        self,
        ctx: wizardParser.ExprContext,
        state: WizardInterpreterState,
        typ: Optional[Type[ValueType]] = None,
    ):
        """
        Evaluate the given expression.

        IMPORTANT: This method can modify the state.

        Args:
            ctx: The expression context to evaluate.
            state: The list of state (can be modified).
            typ: The type of the value to return.

        Returns:
            The result of the given expression.
        """

        if ctx.exception:
            raise WizardParseError(ctx, ctx.exception)

        try:
            value = getattr(self, "visit" + type(ctx).__name__[:-7])(ctx, state)
        except TypeError as te:
            raise WizardTypeError(ctx, *te.args)
        except IndexError as ie:
            raise WizardIndexError(ctx, *ie.args)

        if typ is not None and not isinstance(value.value, typ):
            raise WizardTypeError(
                ctx,
                f"Expected value of type {VariableType.from_pytype(typ)} but got value"
                f" of type {value.type}.",
            )

        return value  # type: ignore



# -*- encoding: utf-8 -*-

from setuptools import find_packages, setup

dev_requires = [
    "black",
    "bump2version",
    "flake8-black",
    "flake8",
    "isort",
    "mypy",
    "types-chardet",
]

setup(
    # Name of the package:
    name="bain-wizard-interpreter",
    # Version of the package:
    version="1.0.2",
    # Find the package automatically (include everything):
    packages=find_packages(exclude=("tests",)),
    package_data={"wizard": ["py.typed"]},
    # Author information:
    author="Holt59",
    author_email="capelle.mikael@gmail.com",
    # Description of the package:
    description="A BAIN Wizard Interpreter based on wizparse.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    # URL for sources:
    url="https://github.com/Holt59/bain-wizard-interpreter",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    # License:
    license="MIT",
    # Requirements:
    install_requires=[
        "chardet",
        "antlr4-python3-runtime==4.10",
    ],
    extras_require={"dev": dev_requires, "tests": ["pytest"]},
)



# -*- encoding: utf-8 -*-

from .contexts import (  # noqa: F401
    WizardAssignmentContext,
    WizardBodyContext,
    WizardBreakContext,
    WizardCancelContext,
    WizardCaseContext,
    WizardCompoundAssignmentContext,
    WizardContinueContext,
    WizardForLoopContext,
    WizardIfContext,
    WizardInterpreterContext,
    WizardReturnContext,
    WizardSelectCasesContext,
    WizardSelectContext,
    WizardSelectManyContext,
    WizardSelectOneContext,
    WizardTerminationContext,
    WizardTopLevelContext,
    WizardWhileLoopContext,
)
from .factory import WizardInterpreterContextFactory  # noqa: F401
from .keywords import (  # noqa: F401
    WizardDeSelectAllContext,
    WizardDeSelectAllPluginsContext,
    WizardDeSelectPluginContext,
    WizardDeSelectSubPackageContext,
    WizardKeywordContext,
    WizardNoteContext,
    WizardRenamePluginContext,
    WizardRequireVersionsContext,
    WizardResetAllPluginNamesContext,
    WizardResetPluginNameContext,
    WizardSelectAllContext,
    WizardSelectAllPluginsContext,
    WizardSelectPluginContext,
    WizardSelectSubPackageContext,
)



# -*- encoding: utf-8 -*-

import copy
from typing import List, Optional, TypeVar, Union

from ..antlr4.wizardParser import wizardParser
from ..expr import WizardExpressionVisitor
from ..keywords import WizardKeywordVisitor
from ..manager import SelectOption
from ..severity import SeverityContext
from ..state import WizardInterpreterState
from .contexts import (
    WizardAssignmentContext,
    WizardBodyContext,
    WizardBreakContext,
    WizardCancelContext,
    WizardCaseContext,
    WizardCompoundAssignmentContext,
    WizardContinueContext,
    WizardExecContext,
    WizardForLoopContext,
    WizardIfContext,
    WizardInterpreterContext,
    WizardReturnContext,
    WizardSelectCasesContext,
    WizardSelectContext,
    WizardSelectManyContext,
    WizardSelectOneContext,
    WizardTerminationContext,
    WizardWhileLoopContext,
)
from .keywords import WizardKeywordContext, make_keyword_context

WizardContext = TypeVar("WizardContext", bound=WizardInterpreterContext)


class WizardInterpreterContextFactory:

    # The visitors:
    _evisitor: WizardExpressionVisitor
    _kvisitor: WizardKeywordVisitor

    # The severity context:
    _severity: SeverityContext

    def __init__(
        self,
        evisitor: WizardExpressionVisitor,
        kvisitor: WizardKeywordVisitor,
        severity: SeverityContext,
    ):
        self._evisitor = evisitor
        self._kvisitor = kvisitor
        self._severity = severity

    @property
    def evisitor(self) -> WizardExpressionVisitor:
        return self._evisitor

    @property
    def kvisitor(self) -> WizardKeywordVisitor:
        return self._kvisitor

    @property
    def severity(self) -> SeverityContext:
        return self._severity

    def _copy_context(
        self,
        context: WizardContext,
        state: Optional[WizardInterpreterState] = None,
    ) -> WizardContext:
        """
        Copy the given context. By default, this method simply calls copy.copy() on
        the context and than state.copy(). This should be  sufficient in most
        cases.

        Args:
            context: The context to copy.
            state: The state to copy into the new context. If None, a copy of the
                context state will be made (see WizardInterpreterState.copy()).

        Returns:
            A copy of this context.
        """
        context = copy.copy(context)
        if state is None:
            context._state = context._state.copy()
        else:
            context._state = state.copy()
        return context

    def _copy_parent(
        self,
        context: WizardInterpreterContext,
        state: Optional[WizardInterpreterState] = None,
    ) -> "WizardInterpreterContext":
        """
        Copy the parent of the given context and update its state using either the
        given state or the state of the context.

        Args:
            context: The state to copy the parent from.
            state: The state to copy into the copy of the parent. If None, a copy of
                the context state will be made (see WizardInterpreterState.copy()).

        Returns:
            A copy of the parent of this context.
        """
        if state is None:
            state = context.state
        return self._copy_context(context.parent, state)

    def make_cancel_context(
        self, context: wizardParser.CancelContext, parent: WizardInterpreterContext
    ) -> WizardCancelContext:
        return WizardCancelContext(self, context, parent)

    def make_return_context(
        self, context: wizardParser.CancelContext, parent: WizardInterpreterContext
    ) -> WizardReturnContext:
        return WizardReturnContext(self, context, parent)

    def make_exec_context(
        self,
        context: wizardParser.FunctionCallContext,
        parent: WizardInterpreterContext,
    ) -> WizardExecContext:
        return WizardExecContext(self, context, parent)

    def make_body_context(
        self,
        context: wizardParser.BodyContext,
        parent: WizardInterpreterContext,
    ) -> WizardBodyContext:
        return WizardBodyContext(self, context, parent)

    def make_assignment_context(
        self,
        context: wizardParser.AssignmentContext,
        parent: WizardInterpreterContext,
    ) -> WizardAssignmentContext:
        return WizardAssignmentContext(self, context, parent)

    def make_compound_assignment_context(
        self,
        context: wizardParser.CompoundAssignmentContext,
        parent: WizardInterpreterContext,
    ) -> WizardCompoundAssignmentContext:
        return WizardCompoundAssignmentContext(self, context, parent)

    def make_keyword_context(
        self,
        context: wizardParser.KeywordStmtContext,
        parent: WizardInterpreterContext,
    ) -> WizardKeywordContext:
        return make_keyword_context(self, context, parent)

    def make_break_context(
        self, context: wizardParser.BreakContext, parent: WizardInterpreterContext
    ) -> WizardBreakContext:
        return WizardBreakContext(self, context, parent)

    def make_continue_context(
        self, context: wizardParser.BreakContext, parent: WizardInterpreterContext
    ) -> WizardContinueContext:
        return WizardContinueContext(self, context, parent)

    def make_termination_context(
        self, parent: WizardInterpreterContext
    ) -> WizardTerminationContext:
        return WizardTerminationContext(self, parent.context, parent)

    def make_for_loop_context(
        self, context: wizardParser.ForStmtContext, parent: WizardInterpreterContext
    ) -> WizardForLoopContext:
        return WizardForLoopContext(self, context, parent)

    def make_while_loop_context(
        self, context: wizardParser.WhileStmtContext, parent: WizardInterpreterContext
    ) -> WizardWhileLoopContext:
        return WizardWhileLoopContext(self, context, parent)

    def make_if_context(
        self, context: wizardParser.IfStmtContext, parent: WizardInterpreterContext
    ) -> WizardIfContext:
        return WizardIfContext(self, context, parent)

    def make_select_context(
        self,
        context: wizardParser.SelectStmtContext,
        parent: WizardInterpreterContext,
    ) -> WizardSelectContext:
        if context.selectOne():
            return WizardSelectOneContext(self, context.selectOne(), parent)
        else:
            return WizardSelectManyContext(self, context.selectMany(), parent)

    def make_select_cases_context(
        self,
        options: List[SelectOption],
        context: wizardParser.SelectStmtContext,
        parent: WizardInterpreterContext,
    ) -> WizardSelectCasesContext:
        return WizardSelectCasesContext(self, options, context, parent)

    def make_case_context(
        self,
        context: Union[wizardParser.CaseStmtContext, wizardParser.DefaultStmtContext],
        parent: WizardInterpreterContext,
    ) -> WizardCaseContext:
        return WizardCaseContext(self, context, parent)
